#!/usr/bin/env python
"""
git-flow

.. program:: git flow

.. cmdoption:: -v, --verbose

       Produce more output.

.. cmdoption:: -h, --help

       Print usage, help and information on the available commands.

"""
#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import re
import sys
import argparse
import subprocess as sub

from gitflow.core import GitFlow, info, GitCommandError
from gitflow.util import itersubclasses
from gitflow.exceptions import (GitflowError, AlreadyInitialized,
                                NotInitialized, BranchTypeExistsError,
                                BaseNotOnBranch, NoSuchBranchError,
                                BaseNotAllowed, BranchExistsError,
                                IllegalVersionFormat)
import gitflow.pivotal as pivotal
from gitflow.review import (BranchReview, ReviewNotAcceptedYet,
                            get_feature_ancestor)

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

def die(*texts):
    raise SystemExit('\n'.join(map(str, texts)))

class NotEmpty(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not values:
            raise argparse.ArgumentError(self, 'must not by empty.')
        setattr(namespace, self.dest, values)


class GitFlowCommand(object):
    """
    This is just an empty class to serve as the base class for all command line
    level sub commands.  Since the git-flow loader will auto-detect all
    subclasses, implementing a new subcommand is as easy as subclassing the
    :class:`GitFlowCommand`.
    """
    @classmethod
    def register_parser(cls, parent):
        raise NotImplementedError("Implement this method in a subclass.")

    @staticmethod
    def run(args):
        raise NotImplementedError("Implement this method in a subclass.")


class VersionCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('version', help='Show the version of gitflow.')
        p.set_defaults(func=cls.run)

    @staticmethod
    def run(args):
        from gitflow import __version__
        print(__version__)


class StatusCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('status', help='Show some status.')
        p.set_defaults(func=cls.run)

    @staticmethod
    def run(args):
        gitflow = GitFlow()
        for name, hexsha, is_active_branch in gitflow.status():
            if is_active_branch:
                prefix = '*'
            else:
                prefix = ' '
            info('%s %s: %s' % (prefix, name, hexsha[:7]))


class InitCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('init',
                              help='Initialize a repository for gitflow.')
        p.add_argument('-f', '--force', action='store_true',
                       help='Force reinitialization of the gitflow preferences.')
        p.add_argument('-d', '--defaults', action='store_true',
                       dest='use_defaults',
                       help='Use default branch naming conventions and prefixes.')
        p.set_defaults(func=cls.run)
        return p

    @staticmethod
    def run(args):
        from . import _init
        _init.run_default(args)


class FeatureCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('feature', help='Manage your feature branches.')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_start(sub)
        cls.register_finish(sub)
        cls.register_checkout(sub)
        cls.register_diff(sub)
        cls.register_rebase(sub)

        cls.register_publish(sub)
        cls.register_pull(sub)
        cls.register_track(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='List all existing feature branches '
                              'in the local and optionally remote repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-a', '--all', action='store_true',
                help='List remote feature branches as well.')
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('feature', 'name', use_tagname=False,
                     verbose=args.verbose, include_remote=args.all)

    #- select
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Select a story in PT and create'
            'a new feature branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Do not fetch from origin before performing local operation.')
        p.add_argument('-r', '--for-release',
                help='Base the feature branch on a release branch.')

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()

        # Make sure --for-release matches the requirements.
        if args.for_release is not None:
            prefix = gitflow.get_prefix('release')
            matcher = gitflow._safe_get('gitflow.release.versionmatcher')
            if not re.match(matcher + '$', args.for_release):
                raise IlegalVersionFormatmat(matcher)
            args.base = prefix + args.for_release

        if not args.no_fetch:
            gitflow.origin().fetch()

        [story_id, args.name] = pivotal.prompt_user_to_select_story()
        story = pivotal.Story(story_id)

        # :fixme: Why does the sh-version not require a clean working dir?
        if args.base is None:
            args.base = gitflow.managers['feature'].default_base()
        gitflow.start_transaction('start feature branch %s (from %s)' % \
                (args.name, args.base))
        try:
            # fetch=False because we are already fetching at the beginning.
            branch = gitflow.create('feature', args.name, args.base,
                                    fetch=False)
        except (NotInitialized, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("Could not create feature branch %r" % args.name, e)

        if release_match is not None:
            story.assign_to_release(release_match.group(1))
        story.start()

        print
        print "Summary of actions:"
        print "- A new branch", branch, "was created, based on", args.base
        print "- You are now on branch", branch
        print ""
        print "Now, start committing on your feature. When done, use:"
        print ""
        print "     git flow feature finish", args.name
        print

    #- finish
    @classmethod
    def register_finish(cls, parent):
        p = parent.add_parser('finish', help='Finish a feature branch ' +
            '(with PTintegration).')
        p.set_defaults(func=cls.run_finish)
        #p.add_argument('-F', '--no-fetch', action='store_true', default=True,
        #        help='Fetch from origin before performing local operation.')
        p.add_argument('-r', '--rebase', action='store_true',
                help='Finish branch by rebasing first.')
        #p.add_argument('-k', '--keep', action='store_true', default=True,
        #        help='Keep branch after performing finish.')
        p.add_argument('-D', '--force-delete', action='store_true',
            default=False, help='Force delete feature branch after finish.')
        p.add_argument('-R', '--no-review', action='store_true',
            default=False, help='Do not post a review request.'
                'not just for the last commit.')
        p.add_argument('-P', '--no-push', action='store_true',
            default=False, help='Do not push the develop branch to origin '
            'after merging with the feature branch.')
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_finish(args):
        gitflow = GitFlow()
        repo = gitflow.repo
        git = repo.git

        f_prefix = gitflow.get_prefix('feature')
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        full_name = f_prefix + name

        #+++ PT story stuff
        sys.stdout.write('Getting data from Pivotal Tracker ... ')
        story_id = pivotal.get_story_id_from_branch_name(name)
        story = pivotal.Story(story_id)
        release = story.get_release()
        print 'OK'

        # Decide on the upstream branch.
        if release:
            # Merge into the release branch.
            r_prefix = gitflow.get_prefix('release')
            short_name = gitflow.nameprefix_or_current('release', release)
            upstream = r_prefix + short_name
        else:
            # Merge into the develop branch.
            upstream = gitflow.develop_name()

        #+++ Git manipulation
        sys.stdout.write('Finishing feature branch ... upstream %s ... ' \
                         % upstream)
        gitflow.finish('feature', name, upstream=upstream,
                       fetch=True, rebase=args.rebase,
                       keep=True, force_delete=args.force_delete,
                       tagging_info=None, push=False)
        print 'OK'

        # Get and set the state of the feature.
        sys.stdout.write('Updating Pivotal Tracker ... ')
        if story.is_finished():
            sys.stdout.write('story already finished, skipping ... ')
        else:
            sys.stdout.write('finishing %s ... ' % story.get_id())
            story.finish()
        print 'OK'

        #+++ Review Request
        if not args.no_review:
            sys.stdout.write('Posting review ... upstream %s ... ' % upstream)
            rev_range = [get_feature_ancestor(full_name),
                         repo.commit(full_name).hexsha]
            desc_cmd = ['git', 'log',
                        "--pretty="
                            "--------------------%n"
                            "Author:   %an <%ae>%n"
                            "Comitter: %cn <%ce>%n"
                            "%n"
                            "%s%n%n"
                            "%b",
                        '{0[0]}...{0[1]}'.format(rev_range)]
            desc = '> Story being reviewed: {0}\n'.format(story.get_url()) + \
                   '\n' \
                   'COMMIT LOG\n' \
                    + sub.check_output(desc_cmd)
            # 7 is the magical offset to get the first commit subject
            summary = desc.split('\n')[7]
            review = BranchReview.from_identifier('feature', name, rev_range)
            review.post(summary, desc)
            print 'OK'

            sys.stdout.write('Posting code review url into Pivotal Tracker ... ')
            comment = 'New patch was uploaded into Review Board: ' + review.get_url()
            story.add_comment(comment)
            print 'OK'

        #+++ Git modify merge message
        sys.stdout.write('Amending merge commit message to include links ... ')
        msg  = 'Finished {0} {1}\n\n'.format(f_prefix, name)
        msg += 'PT-Story-URL: {0}\n'.format(story.get_url())
        msg += 'RB-Review-Request-URL: {0}\n'.format(review.get_url())
        git.commit('--amend', '-m', msg)
        if not args.no_push:
            git.push(gitflow.origin_name(), upstream)
        print 'OK'

    #- checkout
    @classmethod
    def register_checkout(cls, parent):
        p = parent.add_parser('checkout',
                help='Check out (switch to) the given feature branch.')
        p.set_defaults(func=cls.run_checkout)
        p.add_argument('nameprefix', action=NotEmpty)

    @staticmethod
    def run_checkout(args):
        gitflow = GitFlow()
        # NB: Does not default to the current branch as `nameprefix` is required
        branch = gitflow.checkout('feature', args.nameprefix)
        print 'Checking out feature {0}.'.format(branch.name)

    #- diff
    @classmethod
    def register_diff(cls, parent):
        p = parent.add_parser('diff',
                help='Show a diff of changes since this feature branched off.')
        p.set_defaults(func=cls.run_diff)
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_diff(args):
        gitflow = GitFlow()
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        gitflow.start_transaction('diff for feature branch %s' % name)
        gitflow.diff('feature', name)

    #- rebase
    @classmethod
    def register_rebase(cls, parent):
        p = parent.add_parser('rebase',
                help='Rebase a feature branch on top of develop.')
        p.set_defaults(func=cls.run_rebase)
        p.add_argument('-i', '--interactive', action='store_true',
                help='Start an interactive rebase.')
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_rebase(args):
        gitflow = GitFlow()
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        gitflow.start_transaction('rebasing feature branch %s' % name)
        gitflow.rebase('feature', name, args.interactive)

    #- publish
    @classmethod
    def register_publish(cls, parent):
        p = parent.add_parser('publish',
                help='Publish this feature branch to origin.')
        p.set_defaults(func=cls.run_publish)
        p.add_argument('nameprefix', nargs='?')

    @staticmethod
    def run_publish(args):
        gitflow = GitFlow()
        name = gitflow.nameprefix_or_current('feature', args.nameprefix)
        gitflow.start_transaction('publishing feature branch %s' % name)
        branch = gitflow.publish('feature', name)
        print
        print "Summary of actions:"
        print "- A new remote branch '%s' was created" % branch
        print "- The local branch '%s' was configured to track the remote branch" % branch
        print "- You are now on branch '%s'" % branch
        print

    #- pull
    @classmethod
    def register_pull(cls, parent):
        p = parent.add_parser('pull',
                help='Pull a feature branch from a remote peer.')
        p.set_defaults(func=cls.run_pull)
        p.add_argument('remote', action=NotEmpty,
                       help="Remote repository to pull from.")
        p.add_argument('name', nargs='?',
                help='Name of the feature branch to pull. '
                'Defaults to the current branch, if it is a feature branch.')
        # :todo: implement --prefix
        #p.add-argument('-p', '--prefix',
        #               help='Alternative remote feature branch name prefix.')

    @staticmethod
    def run_pull(args):
        gitflow = GitFlow()
        name = gitflow.name_or_current('feature', args.name, must_exist=False)
        gitflow.start_transaction('pulling remote feature branch %s '
                                  'into local banch %s' % (args.remote, name))
        gitflow.pull('feature', args.remote, name)

    #- track
    @classmethod
    def register_track(cls, parent):
        p = parent.add_parser('track',
                help='Track a feature branch from origin.')
        p.set_defaults(func=cls.run_track)
        p.add_argument('name', action=NotEmpty)

    @staticmethod
    def run_track(args):
        gitflow = GitFlow()
        # NB: `args.name` is required since the branch must not yet exist
        gitflow.start_transaction('tracking remote feature branch %s'
                                  % args.name)
        branch = gitflow.track('feature', args.name)
        print
        print "Summary of actions:"
        print "- A new remote tracking branch '%s' was created" % branch
        print "- You are now on branch '%s'" % branch
        print


class ReleaseCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('release', help='Manage your release branches.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_list_stories(sub)
        cls.register_start(sub)
        cls.register_finish(sub)
        cls.register_publish(sub)
        cls.register_track(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='Lists all existing release branches '
                              'in the local repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('release', 'version', use_tagname=True,
                     verbose=args.verbose)

    @classmethod
    def register_list_stories(cls, parent):
        p = parent.add_parser('list_stories',
                              help='Lists all stories that are going to be'
                              'released in the release of your choice.')
        p.set_defaults(func=cls.run_list_stories)
        p.add_argument('--version')

    @staticmethod
    def run_list_stories(args):
        print
        if args.version is None:
            pivotal.Release.dump_all_releases()
        else:
            release = pivotal.Release(args.version)
            release.dump_stories()

    #- start
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Start a new release branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--fetch', action='store_true',
                help='Fetch from origin before performing local operation.')
        p.add_argument('version', action=NotEmpty)

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()
        base = gitflow.develop_name()

        #+ Pivotal Tracker modifications.
        release = pivotal.Release(args.version)
        any_assigned = False
        print
        print 'Stories already assigned to this release:'
        for story in release.iter_stories():
            sys.stdout.write('    ')
            story.dump()
            any_assigned = True
        if not any_assigned:
            print '    None'
        print
        any_candidate = False
        print 'Stories to be newly assigned to this release:'
        for story in release.iter_candidates():
            sys.stdout.write('    ')
            story.dump()
            any_candidate = True
        if not any_candidate:
            print '    None'
        print

        if not any_assigned and not any_candidate:
            raise SystemExit('No stories to be released, aborting...')

        if not release.prompt_for_confirmation():
            raise SystemExit('Aborting...')

        #+ Git modifications.
        sys.stdout.write('Creating release branch (base being %s) ... ' \
                         % base)
        try:
            branch = gitflow.create('release', args.version, base,
                                    fetch=args.fetch)
        except BranchExistsError:
            sys.stdout.write('branch already exists ... ')
        except (NotInitialized, BranchTypeExistsError, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("could not create release branch %r" % args.version, e)
        print 'OK'

        release.start()

        print
        print "Follow-up actions:"
        print "- Bump the version number now!"
        print "- Start committing last-minute fixes in preparing your release"
        print "- When done, run:"
        print
        print "     git flow release finish", args.version
        print

    #- finish
    @classmethod
    def register_finish(cls, parent):
        p = parent.add_parser('finish', help='Finish a release branch.')
        p.set_defaults(func=cls.run_finish)
        # fetch by default
        p.add_argument('-F', '--no-fetch', action='store_true',
                help='Fetch from origin before performing local operation.')
        # push by default
        p.add_argument('-P', '--no-push', action='store_true',
                       #:todo: get "origin" from config
                       help="Push to origin after performing finish.")
        p.add_argument('-k', '--keep', action='store_true',
                help='Keep branch after performing finish.')
        p.add_argument('-R', '--ignore-missing-reviews', action='store_true',
                       help='Just print a warning if there is no review for '
                            'a feature branch that is assigned to this release,'
                            ' do not fail.')
        p.add_argument('version', nargs='?')

        g = p.add_argument_group('tagging options')
        g.add_argument('-n', '--notag', action='store_true',
                       help="Don't tag this release.")
        g.add_argument('-m', '--message',
                       help="Use the given tag message.")
        g.add_argument('-s', '--sign', action='store_true',
                help="Sign the release tag cryptographically.")
        g.add_argument('-u', '--signingkey',
                help="Use the given GPG-key for the digital signature "
                     "instead of the default git uses (implies -s).")

    @staticmethod
    def run_finish(args):
        gitflow = GitFlow()
        git     = gitflow.git
        origin  = gitflow.origin()
        version = gitflow.name_or_current('release', args.version)

        #+++ Check QA
        release = pivotal.Release(version)
        print "Checking if all relevant stories have been QA'd ... "
        try:
            release.try_deliver()
        except GitflowError:
            raise SystemExit('FAIL')
        print 'OK'

        #+++ Close (submit) all relevant reviews in Review Board
        print 'Submitting all relevant review requests ... '
        feature_prefix = gitflow.get_prefix('feature')
        err = None
        for story in release.iter_stories():
            prefix = feature_prefix + str(story.get_id())
            try:
                r = BranchReview.from_prefix(prefix)
            except NoSuchBranchError, e:
                err = e
                print '    ' + str(e)
                continue
            r.submit()
            print '    ' + str(r.get_id())
        if not args.ignore_missing_reviews and err is not None:
            raise SystemExit('FAIL')
        print 'OK'

        #+++ Merge release branch into develop and master
        sys.stdout.write('Finishing release branch %s ... ' % version)
        tagging_info = None
        if not args.notag:
            tagging_info = dict(
                sign=args.sign or args.signingkey,
                signingkey=args.signingkey,
                message=args.message)
        gitflow.finish('release', version,
                                 fetch=(not args.no_fetch), rebase=False,
                                 keep=args.keep, force_delete=False,
                                 tagging_info=tagging_info, push=(not args.no_push))
        print 'OK'

        #+++ Collect local and remote branches to be deleted
        sys.stdout.write('Collecting branches to be deleted ... ')
        local_branches  = list()
        remote_branches = list()

        #+ Collect features to be deleted.
        origin_prefix = str(origin) + '/'
        feature_prefix = gitflow.get_prefix('feature')
        # refs = [<type>/<id>/...]
        refs = [str(ref)[len(origin_prefix):] for ref in origin.refs]
        for story in release.iter_stories():
            # prefix = <feature-prefix>/<id>
            prefix = feature_prefix + str(story.get_id())
            try:
                name = gitflow.nameprefix_or_current('feature', prefix)
                local_branches.append(feature_prefix + name)
            except NoSuchBranchError:
                pass
            for ref in refs:
                # if <feature-prefix>/... startswith <feature-prefix>/<id>
                if ref.startswith(prefix):
                    remote_branches.append(ref)
        #+ Collect releases to be deleted.
        release_branch = gitflow.get_prefix('release') + version
        try:
            gitflow.nameprefix_or_current('release', version)
            local_branches.append(release_branch)
        except NoSuchBranchError:
            pass
        if release_branch in refs:
            remote_branches.append(release_branch)
        print 'OK'

        #+++ Delete local and remote branches that are a part of this release
        sys.stdout.write('Checking out %s ... ' % gitflow.develop_name())
        git.checkout(gitflow.develop_name())
        print 'OK'
        #+ Delete local branches.
        print 'Deleting local branches ...'
        for branch in local_branches:
            git.branch('-D', branch)
            print '    ' + branch
        print '    OK'
        #+ Delete remote branches.
        print 'Deleting remote branches ...'
        for branch in remote_branches:
            print '    ' + branch
        refspecs = [(':' + b) for b in remote_branches]
        git.push(str(origin), *refspecs)
        print '    OK'

        #+++ Deliver all relevant stories in Pivotal Tracker
        release.deliver()

    #- publish
    @classmethod
    def register_publish(cls, parent):
        p = parent.add_parser('publish',
                help='Publish this release branch to origin.')
        p.set_defaults(func=cls.run_publish)
        p.add_argument('version', nargs='?')

    @staticmethod
    def run_publish(args):
        gitflow = GitFlow()
        version = gitflow.name_or_current('release', args.version)
        gitflow.start_transaction('publishing release branch %s' % version)
        branch = gitflow.publish('release', version)
        print
        print "Summary of actions:"
        print "- A new remote branch '%s' was created" % branch
        print "- The local branch '%s' was configured to track the remote branch" % branch
        print "- You are now on branch '%s'" % branch
        print

    #- track
    @classmethod
    def register_track(cls, parent):
        p = parent.add_parser('track',
                help='Track a release branch from origin.')
        p.set_defaults(func=cls.run_track)
        p.add_argument('version', action=NotEmpty)

    @staticmethod
    def run_track(args):
        gitflow = GitFlow()
        # NB: `args.version` is required since the branch must not yet exist
        gitflow.start_transaction('tracking remote release branch %s'
                                  % args.version)
        branch = gitflow.track('release', args.version)
        print
        print "Summary of actions:"
        print "- A new remote tracking branch '%s' was created" % branch
        print "- You are now on branch '%s'" % branch
        print


class HotfixCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('hotfix', help='Manage your hotfix branches.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_start(sub)
        cls.register_finish(sub)
        cls.register_publish(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='Lists all existing hotfix branches '
                              'in the local repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('hotfix', 'version', use_tagname=True,
                     verbose=args.verbose)

    #- start
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Start a new hotfix branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--fetch', action='store_true',
                       #:todo: get "origin" from config
                help='Fetch from origin before performing local operation.')
        p.add_argument('version', action=NotEmpty)
        p.add_argument('base', nargs='?')

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()
        # NB: `args.version` is required since the branch must not yet exist
        # :fixme: get default value for `base`
        gitflow.start_transaction('create hotfix branch %s (from %s)' % \
                (args.version, args.base))
        try:
            branch = gitflow.create('hotfix', args.version, args.base,
                                    fetch=args.fetch)
        except (NotInitialized, BranchTypeExistsError, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("Could not create hotfix branch %r" % args.version, e)
        print
        print "Summary of actions:"
        print "- A new branch", branch, "was created, based on", args.base
        print "- You are now on branch", branch
        print ""
        print "Follow-up actions:"
        print "- Bump the version number now!"
        print "- Start committing your hot fixes"
        print "- When done, run:"
        print
        print "     git flow hotfix finish", args.version

    #- finish
    @classmethod
    def register_finish(cls, parent):
        p = parent.add_parser('finish', help='Finish a hotfix branch.')
        p.set_defaults(func=cls.run_finish)
        p.add_argument('-F', '--fetch', action='store_true',
                help='Fetch from origin before performing local operation.')
        p.add_argument('-p', '--push', action='store_true',
                       #:todo: get "origin" from config
                       help="Push to origin after performing finish.")
        p.add_argument('-k', '--keep', action='store_true',
                help='Keep branch after performing finish.')
        p.add_argument('version', nargs='?')

        g = p.add_argument_group('tagging options')
        g.add_argument('-n', '--notag', action='store_true',
                       help="Don't tag this hotfix.")
        g.add_argument('-m', '--message',
                       help="Use the given tag message.")
        g.add_argument('-s', '--sign', action='store_true',
                help="Sign the hotfix tag cryptographically.")
        g.add_argument('-u', '--signingkey',
                help="Use this given GPG-key for the digital signature "
                     "instead of the default git uses (implies -s).")

    @staticmethod
    def run_finish(args):
        gitflow = GitFlow()
        version = gitflow.name_or_current('hotfix', args.version)
        gitflow.start_transaction('finishing hotfix branch %s' % version)
        tagging_info = None
        if not args.notag:
            tagging_info = dict(
                sign=args.sign or args.signingkey,
                signingkey=args.signingkey,
                message=args.message)
        release = gitflow.finish('hotfix', version,
                                 fetch=args.fetch, rebase=False,
                                 keep=args.keep, force_delete=False,
                                 tagging_info=tagging_info)

    #- publish
    @classmethod
    def register_publish(cls, parent):
        p = parent.add_parser('publish',
                help='Publish this hotfix branch to origin.')
        p.set_defaults(func=cls.run_publish)
        p.add_argument('version', nargs='?')

    @staticmethod
    def run_publish(args):
        gitflow = GitFlow()
        version = gitflow.name_or_current('hotfix', args.version)
        gitflow.start_transaction('publishing hotfix branch %s' % version)
        branch = gitflow.publish('hotfix', version)
        print
        print "Summary of actions:"
        print "- A new remote branch '%s' was created" % branch
        print "- The local branch '%s' was configured to track the remote branch" % branch
        print "- You are now on branch '%s'" % branch
        print


class SupportCommand(GitFlowCommand):
    @classmethod
    def register_parser(cls, parent):
        p = parent.add_parser('support', help='Manage your support branches.')
        p.add_argument('-v', '--verbose', action='store_true',
           help='Be verbose (more output).')
        sub = p.add_subparsers(title='Actions')
        cls.register_list(sub)
        cls.register_start(sub)

    #- list
    @classmethod
    def register_list(cls, parent):
        p = parent.add_parser('list',
                              help='Lists all existing support branches '
                              'in the local repository.')
        p.set_defaults(func=cls.run_list)
        p.add_argument('-v', '--verbose', action='store_true',
                help='Be verbose (more output).')

    @staticmethod
    def run_list(args):
        gitflow = GitFlow()
        gitflow.start_transaction()
        gitflow.list('support', 'version', use_tagname=True,
                     verbose=args.verbose)

    #- start
    @classmethod
    def register_start(cls, parent):
        p = parent.add_parser('start', help='Start a new support branch.')
        p.set_defaults(func=cls.run_start)
        p.add_argument('-F', '--fetch', action='store_true',
                help='Fetch from origin before performing local operation.')
        p.add_argument('name', action=NotEmpty)
        p.add_argument('base', nargs='?')

    @staticmethod
    def run_start(args):
        gitflow = GitFlow()
        # NB: `args.name` is required since the branch must not yet exist
        # :fixme: get default value for `base`
        gitflow.start_transaction('create support branch %s (from %s)' %
                (args.name, args.base))
        try:
            branch = gitflow.create('support', args.name, args.base,
                                    fetch=args.fetch)
        except (NotInitialized, BranchTypeExistsError, BaseNotOnBranch):
            # printed in main()
            raise
        except Exception, e:
            die("Could not create support branch %r" % args.name, e)
        print
        print "Summary of actions:"
        print "- A new branch", branch, "was created, based on", args.base
        print "- You are now on branch", branch
        print ""


def main():
    parser = argparse.ArgumentParser(prog='git flow')
    placeholder = parser.add_subparsers(title='Subcommands')
    for cls in itersubclasses(GitFlowCommand):
        cls.register_parser(placeholder)
    args = parser.parse_args()
    try:
        args.func(args)
    except KeyboardInterrupt:
        raise SystemExit('Aborted by user request.')


if __name__ == '__main__':
    try:
        main()
    except (GitflowError, GitCommandError), e:
        raise SystemExit('Error: %s' %e)



#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

import os
import sys
import time
import datetime
import ConfigParser
from functools import wraps

import git
from git import (Git, Repo, InvalidGitRepositoryError, RemoteReference,
                 GitCommandError)

from gitflow.branches import BranchManager
from gitflow.util import itersubclasses

from gitflow.exceptions import (NotInitialized, BranchExistsError,
                                BranchTypeExistsError, MergeConflict,
                                NoSuchRemoteError, NoSuchBranchError,
                                Usage, BadObjectError)

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

def datetime_to_timestamp(d):
    return time.mktime(d.timetuple()) + d.microsecond / 1e6


def requires_repo(f):
    @wraps(f)
    def _inner(self, *args, **kwargs):
        if self.repo is None:
            msg = 'This repo has not yet been initialized for git-flow.'
            raise NotInitialized(msg)
        return f(self, *args, **kwargs)
    return _inner

def requires_initialized(f):
    @wraps(f)
    def _inner(self, *args, **kwargs):
        msg = 'This repo has not yet been initialized for git-flow.'
        if (not self.is_initialized() or
            not self.master_name() in self.repo.branches or
            not self.develop_name() in self.repo.branches):
            raise NotInitialized(msg)
        try:
            self.get('reviewboard.url')
            self.get('reviewboard.server')
            self.get('gitflow.rb.repoid')
            self.get('gitflow.pt.token')
            self.get('gitflow.pt.projectid')
        except Exception:
            raise NotInitialized(msg)
        return f(self, *args, **kwargs)
    return _inner


def info(*texts):
    for txt in texts:
        print txt

def warn(*texts):
    for txt in texts:
        print >> sys.stderr, txt


class _NONE:
    pass


class GitFlow(object):
    """
    Creates a :class:`GitFlow` instance.

    :param working_dir:
        The directory where the Git repo is located.  If not specified, the
        current working directory is used.

    When a :class:`GitFlow` class is instantiated, it auto-discovers all
    subclasses of :class:`gitflow.branches.BranchManager`, so there is no
    explicit registration required.
    """

    def _discover_branch_managers(self):
        managers = {}
        for cls in itersubclasses(BranchManager):
            # TODO: Initialize managers with the gitflow branch prefixes
            managers[cls.identifier] = cls(self)
        return managers

    def __init__(self, working_dir='.'):
        # Allow Repos to be passed in instead of strings
        self.repo = None
        if isinstance(working_dir, Repo):
            self.working_dir = working_dir.working_dir
        else:
            self.working_dir = working_dir

        self.git = Git(self.working_dir)
        try:
            self.repo = Repo(self.working_dir)
        except InvalidGitRepositoryError:
            pass

        self.managers = self._discover_branch_managers()
        self.defaults = {
            'gitflow.branch.master': 'master',
            'gitflow.branch.develop': 'develop',
            'gitflow.prefix.versiontag': '',
            'gitflow.origin': 'origin',
            'gitflow.release.versionmatcher': '[0-9]+([.][0-9]+){2}',
            }
        for identifier, manager in self.managers.items():
            self.defaults['gitflow.prefix.%s' % identifier] = manager.DEFAULT_PREFIX


    def _init_config(self, master=None, develop=None, prefixes={}, names={},
                     force_defaults=False):
        for setting, default in self.defaults.items():
            if force_defaults:
                value = default
            elif setting == 'gitflow.branch.master':
                value = master
            elif setting == 'gitflow.branch.develop':
                value = develop
            elif setting.startswith('gitflow.prefix.'):
                name = setting[len('gitflow.prefix.'):]
                value = prefixes.get(name, None)
            else:
                name = setting[len('gitflow.'):]
                value = names.get(name, None)
            if value is None:
                value = self.get(setting, default)
            self.set(setting, value)

    def _init_initial_commit(self):
        master = self.master_name()
        if master in self.repo.branches:
            # local `master` branch exists
            return
        elif self.origin_name(master) in self.repo.refs:
            # the origin branch counterpart exists
            origin = self.repo.refs[self.origin_name(master)]
            branch = self.repo.create_head(master, origin)
            branch.set_tracking_branch(origin)
        elif self.repo.heads:
            raise NotImplementedError('Local and remote branches exist, '
                                     'but neither %s nor %s'
                                     % (master, self.origin_name(master) ))
        else:
            # Create 'master' branch
            info('Creating branch %r' % master)
            c = self.repo.index.commit('Initial commit', head=False)
            self.repo.create_head(master, c)

    def _init_develop_branch(self):
        # assert master already exists
        assert self.master_name() in self.repo.refs
        develop = self.develop_name()
        if develop in self.repo.branches:
            # local `develop` branch exists, but do not switch there
            return
        if self.origin_name(develop) in self.repo.refs:
            # the origin branch counterpart exists
            origin = self.repo.refs[self.origin_name(develop)]
            branch = self.repo.create_head(develop, origin)
            branch.set_tracking_branch(origin)
        else:
            # Create 'develop' branch
            info('Creating branch %r' % develop)
            branch = self.repo.create_head(develop, self.master())
        # switch to develop branch if its newly created
        info('Switching to branch %s' % branch)
        branch.checkout()


    def _enforce_git_repo(self):
        """
        Ensure a (maybe empty) repository exists we can work on.

        This is to be used by the `init` sub-command.
        """
        if self.repo is None:
            self.git.init(self.working_dir)
            self.repo = Repo(self.working_dir)


    def _enforce_services(self):
        err = False
        for option in ('gitflow.pt.token',
                       'reviewboard.url',
                       'reviewboard.server'):
            try:
                value = self.get(option)
                if option == 'reviewboard.url' and not value.endswith('/'):
                    err = True
                    print """
Git config key 'reviewboard.url' must contain a trailing slash.
Update your configuration by executing

    $ git config [--global] reviewboard.url %s
""" % (value + '/')
            except Exception:
                if option == 'gitflow.pt.token':
                    value = self.get('workflow.token')
                    self.set('gitflow.pt.token', value)
                    continue
                err = True
                print """
Git config '%s' missing, please fill it in by executing

    $ git config [--global] %s <value>
""" % (option, option)

        if err:
            raise SystemExit('Operation canceled.')


    def init(self, master=None, develop=None, prefixes={}, names={},
             force_defaults=False):
        self._enforce_git_repo()
        self._init_config(master, develop, prefixes, names, force_defaults)
        self._init_initial_commit()
        self._init_develop_branch()
        return self

    def is_initialized(self):
        return (self.repo and
                self.is_set('gitflow.branch.master') and
                self.is_set('gitflow.branch.develop') and
                self.is_set('gitflow.prefix.feature') and
                self.is_set('gitflow.prefix.release') and
                self.is_set('gitflow.prefix.hotfix') and
                self.is_set('gitflow.prefix.support') and
                self.is_set('gitflow.prefix.versiontag') and
                self.is_set('gitflow.release.versionmatcher'))

    def _parse_setting(self, setting):
        groups = setting.split('.', 2)
        if len(groups) == 2:
            section, option = groups
        elif len(groups) == 3:
            section, subsection, option = groups
            section = '%s "%s"' % (section, subsection)
        else:
            raise ValueError('Invalid setting name: %s' % setting)
        return (section, option)

    @requires_repo
    def get(self, setting, default=_NONE):
        section, option = self._parse_setting(setting)
        try:
            return self.repo.config_reader().get_value(section, option)
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            if default is not _NONE:
                return default
            raise

    def get_prefix(self, identifier):
        return self._safe_get('gitflow.prefix.%s' % (identifier,))

    @requires_repo
    def set(self, setting, value):
        section, option = self._parse_setting(setting)
        self.repo.config_writer().set_value(section, option, value)

    def is_set(self, setting):
        return self.get(setting, None) is not None


    @requires_repo
    def _safe_get(self, setting_name):
        try:
            return self.get(setting_name)
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            raise NotInitialized('This repo has not yet been initialized.')

    def master_name(self):
        return self._safe_get('gitflow.branch.master')

    def develop_name(self):
        return self._safe_get('gitflow.branch.develop')

    def origin_name(self, name=None):
        origin = self.get('gitflow.origin', self.defaults['gitflow.origin'])
        if name is not None:
            return origin + '/' + name
        else:
            return origin

    @requires_repo
    def require_remote(self, name):
        try:
            return self.repo.remotes[name]
        except IndexError:
            raise NoSuchRemoteError("Remote '{0}' was not found.".format(name))

    def origin(self):
        return self.require_remote(self.origin_name())

    @requires_repo
    def develop(self):
        return self.repo.branches[self.develop_name()]

    @requires_repo
    def master(self):
        return self.repo.branches[self.master_name()]

    @requires_repo
    def branch_names(self, remote=False):
        if remote:
            return [r.name
                    for r in self.repo.refs
                    if isinstance(r, RemoteReference)]
        else:
            return [r.name for r in self.repo.branches]

    @requires_repo
    def nameprefix_or_current(self, identifier, prefix):
        """
        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param prefix: If the empty, see if the current branch is of
            type `identifier`. If so, returns the current branches
            short name, otherwise raises :exc:`NoSuchBranchError`.

        If exactly one branch of type `identifier` starts with
        the given name `prefix`, returns that branches short name.
        Raises :exc:`NoSuchBranchError` in case no branch exists with
        the given prefix, or :exc:`PrefixNotUniqueError` in case
        multiple matches are found.
        """
        repo = self.repo
        manager = self.managers[identifier]
        if not prefix:
            if repo.active_branch.name.startswith(manager.prefix):
                return manager.shorten(repo.active_branch.name)
            else:
                raise NoSuchBranchError('The current branch is no %s branch. '
                    'Please specify one explicitly.' % identifier)
        return manager.shorten(manager.by_name_prefix(prefix).name)

    @requires_repo
    def name_or_current(self, identifier, name, must_exist=True):
        """
        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param  name:
           If the `name` is empty, see if the current branch is of
           type `identifier`. If so, returns the current branches
           short name, otherwise raises :exc:`NoSuchBranchError`.

        :param must_exist: If `True` (the default), raises
            :exc:`NoSuchBranchError` in case no branch exists with the
            given `name`.

        Otherwise return the `name` unchanged.
        """
        repo = self.repo
        manager = self.managers[identifier]
        if not name:
            if repo.active_branch.name.startswith(manager.prefix):
                return manager.shorten(repo.active_branch.name)
            else:
                raise NoSuchBranchError('The current branch is no %s branch. '
                    'Please specify one explicitly.' % identifier)
        elif must_exist and not manager.full_name(name) in (b.name for b in manager.list()):
            raise NoSuchBranchError('There is no %s branch named %s.'
                                    % (identifier, name))
        return name


    @requires_repo
    def status(self):
        result = []
        for b in self.repo.branches:
            tup = self.branch_info(b.name)
            result.append(tup)
        return result

    @requires_repo
    def branch_info(self, name):
        active_branch = self.repo.active_branch
        b = self.repo.heads[name]
        return (name, b.commit.hexsha, b == active_branch)

    @requires_repo
    def is_dirty(self):
        """
        Returns whether or not the current working directory contains
        uncommitted changes.
        """
        return self.repo.is_dirty()

    @requires_repo
    def has_staged_commits(self):
        """
        Returns whether or not the current repo contains local changes
        checked into the index but not committed.
        """
        return len(self.repo.index.diff(self.repo.head.commit)) > 0


    @requires_repo
    def require_no_merge_conflict(self):
        """
        Raises :exc:`MergeConflict` if the current working directory
        contains a merge conflict.
        """
        try:
            git.Reference(self.repo, 'MERGE_HEAD', check_path=False).commit
            # reference exists, so there is a merge conflict
            raise MergeConflict()
        except ValueError:
            # no such reference, so there is no merge conflict
            pass


    def is_merged_into(self, commit, target_branch):
        """
        Checks whether `commit` is successfully merged into branch
        `target_branch`.

        :param commit:
            The commit or branch that ought to be merged. This may be
            a full branch-name, a commit-hexsha or any of branch-,
            head-, reference- or commit-object.

        :param target_branch:
            The branch which should contain the commit. This may be a
            full branch-name, or any of branch-, head- or
            reference-object.
        """
        try:
            commit = self.repo.rev_parse(str(commit))
        except git.BadObject:
            raise BadObjectError(commit)
        if isinstance(target_branch, git.RemoteReference):
            target_branch = 'remotes/' + target_branch.name
        elif isinstance(target_branch, git.SymbolicReference):
            target_branch = target_branch.name
        # :todo: implement this more efficiently
        return target_branch in [
            b.lstrip('* ')
            for b in self.git.branch('-a', '--contains', commit).splitlines()]


    def must_be_uptodate(self, branch, fetch):
        remote_branch = self.origin_name(branch)
        if remote_branch in self.branch_names(remote=True):
            if fetch:
                self.origin().fetch(branch)
            self.require_branches_equal(branch, remote_branch)

    @requires_repo
    def _compare_branches(self, branch1, branch2):
        """
        Tests whether branches and their 'origin' counterparts have
        diverged and need merging first. It returns error codes to
        provide more detail, like so:

        0    Branch heads point to the same commit
        1    First given branch needs fast-forwarding
        2    Second given branch needs fast-forwarding
        3    Branch needs a real merge
        4    There is no merge base, i.e. the branches have no common ancestors
        """
        try:
            commit1 = self.repo.rev_parse(branch1)
            commit2 = self.repo.rev_parse(branch2)
        except git.BadObject, e:
            raise NoSuchBranchError(e.args[0])
        if commit1 == commit2:
            return 0
        try:
            base = self.repo.git.merge_base(commit1, commit2)
        except GitCommandError:
            return 4
        if base == commit1:
            return 1
        elif base == commit2:
            return 2
        else:
            return 3


    @requires_repo
    def require_branches_equal(self, branch1, branch2):
        status = self._compare_branches(branch1, branch2)
        if status == 0:
            # branches are equal
            return
        else:
            warn("Branches '%s' and '%s' have diverged." % (branch1, branch2))
            if status == 1:
                raise SystemExit("And branch '%s' may be fast-forwarded." % branch1)
            elif status == 2:
                # Warn here, since there is no harm in being ahead
                warn("And local branch '%s' is ahead of '%s'." % (branch1, branch2))
            else:
                raise SystemExit("Branches need merging first.")

    @requires_repo
    def start_transaction(self, message=None):
        if message:
            info(message)

    @requires_initialized
    def tag(self, tagname, commit, message=None, sign=False, signingkey=None):
        kwargs = {}
        if sign:
            kwargs['s'] = True
        if signingkey:
            kwargs['u'] = signingkey
        self.repo.create_tag(tagname, commit, message=message or None, **kwargs)

    #
    #====== sub commands =====
    #

    @requires_repo
    def list(self, identifier, arg0_name, verbose, use_tagname,
            include_remote=False):
        """
        List the all branches of the given type. If there are not
        branches of this type, raises :exc:`Usage` with an
        explanation on how to start a branch of this type.

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param arg0_name:
            Name of the first argument for the command line to be put
            into the explanation on how to start a branch of this
            type. This typically is `name` or `version`.

        :param verbose:
            If True, give more information about the state of the
            branch: Whether it's ahead or behind it's default base,
            may be rebased, etc.

        :param use_tagname:
            If True, try to describe the state based on the next tag.
        """
        repo = self.repo
        manager = self.managers[identifier]
        branches = manager.list()
        if include_remote:
            branches += manager.list(remote=True)
        if not branches:
            raise Usage(
                'No %s branches exist.' % identifier,
                'You can start a new %s branch with the command:' % identifier,
                '    git flow %s start <%s> [<base>]' % (identifier, arg0_name)
                )

        # determine the longest branch name
        width = max(len(b.name) for b in branches) - len(manager.prefix) + 1

        basebranch_sha = repo.branches[manager.default_base()].commit.hexsha

        for branch in branches:
            if repo.active_branch == branch:
                prefix = '* '
            else:
                prefix = '  '

            name = manager.shorten(branch.name)
            extra_info = ''

            if verbose:
                name = name.ljust(width)
                branch_sha = branch.commit.hexsha
                base_sha = repo.git.merge_base(basebranch_sha, branch_sha)
                if branch_sha == basebranch_sha:
                    extra_info = '(no commits yet)'
                elif use_tagname:
                    try:
                        extra_info = self.git.name_rev('--tags','--name-only',
                                                       '--no-undefined', base_sha)
                        extra_info = '(based on %s)' % extra_info
                    except GitCommandError:
                        pass
                if not extra_info:
                    if base_sha == branch_sha:
                        extra_info = '(is behind %s, may ff)' % manager.default_base()
                    elif base_sha == basebranch_sha:
                        extra_info = '(based on latest %s)' % manager.default_base()
                    else:
                        extra_info = '(may be rebased)'

            info(prefix + name + extra_info)


    @requires_initialized
    def create(self, identifier, name, base, fetch):
        """
        Creates a branch of the given type, with the given short name.

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to create.

        :param base:
            The alternative base to branch off from.  If not given, the default
            base for the given branch type is used.

        :returns:
            The newly created :class:`git.refs.Head` branch.
        """
        return self.managers[identifier].create(name, base, fetch=fetch)


    @requires_initialized
    def finish(self, identifier, name, fetch, rebase, keep, force_delete,
               tagging_info, push=False, **kwargs):
        """
        Finishes a branch of the given type, with the given short name.

        :param identifier:
            The identifier for the type of branch to finish.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to finish.
        """
        mgr = self.managers[identifier]
        branch = mgr.by_name_prefix(name)
        try:
            self.require_no_merge_conflict()
        except MergeConflict, e:
            raise Usage(e,
                        "You can then complete the finish by running it again:",
                        "    git flow %s finish %s" % (identifier, name)
                        )
        return mgr.finish(mgr.shorten(branch.name), fetch=fetch, rebase=rebase,
                          keep=keep, force_delete=force_delete,
                          tagging_info=tagging_info, push=push, **kwargs)

    @requires_initialized
    def checkout(self, identifier, name):
        """
        Checkout a branch of the given type, with the given short name.

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to create.

        :returns:
            The checked out :class:`git.refs.Head` branch.
        """
        mgr = self.managers[identifier]
        try:
            branch = mgr.by_name_prefix(name)
            return branch.checkout()
        except NoSuchBranchError, ex:
            pass

        branch = mgr.by_name_prefix(name, remote=True)
        self.git.checkout(branch.name[len(self.origin_name())+1:])
        return branch

    @requires_initialized
    def diff(self, identifier, name):
        """
        Print the diff of changes since this branch branched off.

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to create.
        """
        repo = self.repo
        mgr = self.managers[identifier]
        full_name = mgr.full_name(name)
        base = self.git.merge_base(mgr.default_base(), full_name)
        print self.git.diff('%s..%s' % (base, full_name))

    @requires_initialized
    def rebase(self, identifier, name, interactive, upstream=None):
        """
        Rebase a branch of the given type, with the given short name,
        on top of it's default base.

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to create.

        :param interactive:
            If True, do an interactive rebase.
        """
        warn("Will try to rebase %s branch '%s' ..." % (identifier, name))
        repo = self.repo
        mgr = self.managers[identifier]
        full_name = mgr.full_name(name)
        # :todo: require_clean_working_tree
        self.checkout(identifier, name)
        args = []
        if interactive:
            args.append('-i')
        if not upstream:
            upstream = mgr.default_base()
        args.append(upstream)
        self.git.rebase(*args)

    @requires_initialized
    def publish(self, identifier, name):
        """
        Publish a branch of the given type, with the given short name,
        to `origin` (or whatever is configured as `remote` for gitflow.)

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to create.
        """
        repo = self.repo
        mgr = self.managers[identifier]

        # sanity checks
        # :todo: require_clean_working_tree
        full_name = mgr.full_name(name)
        remote_name = self.origin_name(full_name)
        if not full_name in repo.branches:
            raise NoSuchBranchError(full_name)
        if remote_name in repo.refs:
            raise BranchExistsError(remote_name)
        # :todo: check if full_name already has a tracking branch
        # :todo: check if full_name already has the same tracking branch

        # create remote branch
        origin = self.origin()
        info = origin.push('%s:refs/heads/%s' % (full_name, full_name))[0]
        origin.fetch()
        # configure remote tracking
        repo.branches[full_name].set_tracking_branch(info.remote_ref)


    @requires_initialized
    def pull(self, identifier, remote, name):
        """
        Pull a branch of the given type, with the given short name,
        from the given remote peer.

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param remote:
            The remote to pull from. This must have been configured by
            `git remote add ...`.

        :param name:
            The friendly (short) name to create.
        """

        def avoid_accidental_cross_branch_action(branch_name):
            current_branch = repo.active_branch
            if branch_name != current_branch.name:
                warn("Trying to pull from '%s' while currently on branch '%s'."
                     % (branch_name , current_branch))
                raise SystemExit("To avoid unintended merges, git-flow aborted.")

        repo = self.repo
        mgr = self.managers[identifier]
        full_name = mgr.full_name(name)
        # To avoid accidentally merging different feature branches
        # into each other, die if the current feature branch differs
        # from the requested $NAME argument.
        if repo.active_branch.name.startswith(self.get_prefix(identifier)):
            # We are on a local `identifier` branch already, so `full_name`
            # must be equal to the current branch.
            avoid_accidental_cross_branch_action(full_name)
        # :todo: require_clean_working_tree
        if full_name in self.repo.branches:
            # Again, avoid accidental merges
            avoid_accidental_cross_branch_action(full_name)
            # We already have a local branch called like this, so
            # simply pull the remote changes in
            self.require_remote(remote).pull(full_name)
            # :fixme: why is the branch not checked out here?
            info("Pulled %s's changes into %s." % (remote, full_name))
        else:
            # Setup the non-tracking local branch clone for the first time
            self.require_remote(remote).fetch(full_name+':'+full_name)
            repo.heads[full_name].checkout()
            info("Created local branch %s based on %s's %s."
                 % (full_name, remote, full_name))

    @requires_initialized
    def track(self, identifier, name):
        """
        Track a branch of the given type, with the given short name,
        from `origin` (or whatever is configured as `remote` for
        gitflow.)

        :param identifier:
            The identifier for the type of branch to create.
            A :class:`BranchManager <git.branches.BranchManager>` for the given
            identifier must exist in the :attr:`self.managers`.

        :param name:
            The friendly (short) name to create.

        :param base:
            The alternative base to branch off from.  If not given, the default
            base for the given branch type is used.

        :returns:
            The newly created :class:`git.refs.Head` branch.
        """
        repo = self.repo
        mgr = self.managers[identifier]
        # sanity checks
        # :todo: require_clean_working_tree
        full_name = mgr.full_name(name)
        if full_name in repo.branches:
            raise BranchExistsError(full_name)
        self.origin().fetch(full_name)
        remote_branch = self.origin().refs[full_name]
        branch = repo.create_head(full_name, remote_branch)
        branch.set_tracking_branch(remote_branch)
        return branch.checkout()



from gitflow.core import GitFlow
import busyflow.pivotal as pt
import string
from colorama import Style
from colorama import init
init()
import sys
import re
import itertools

from gitflow.exceptions import (NotInitialized, GitflowError,
                                ReleaseAlreadyAssigned, IllegalVersionFormat,
                                StatusError, NoSuchBranchError)


_gitflow = GitFlow()

def _get_client():
    return pt.PivotalClient(token=_gitflow._safe_get('gitflow.pt.token'))

def _get_project_id():
    return _gitflow._safe_get('gitflow.pt.projectid')

def _get_version_matcher():
    return _gitflow._safe_get('gitflow.release.versionmatcher')

def _iter_current_stories():
    client = _get_client()
    pid = _get_project_id()
    iterations = client.iterations.current(pid)
    if 'iterations' not in iterations:
        return
    for iteration in iterations['iterations']:
        for story in iteration['stories']:
            yield Story.from_dict(story)

def _iter_backlog_stories():
    client = _get_client()
    pid = _get_project_id()
    iterations = client.iterations.backlog(pid)
    if 'iterations' not in iterations:
        return
    for iteration in iterations['iterations']:
        for story in iteration['stories']:
            yield Story.from_dict(story)

def _iter_stories():
    return itertools.chain(_iter_current_stories(), _iter_backlog_stories())

def _check_version_format(version):
    matcher = _get_version_matcher()
    if not re.match(matcher + '$', version):
        raise IllegalVersionFormat(matcher)

def list_projects():
    projects = _get_client().projects.all()['projects']
    return [(p['id'], p['name']) for p in projects]


class Story(object):
    def __init__(self, story_id, _skip_story_download=False):
        self._project_id = _get_project_id()
        self._client = _get_client()
        if _skip_story_download:
            return
        payload = self._client.stories.get(self._project_id, story_id)
        self._story = payload['story']

    def get_id(self):
        return self._story['id']

    def get_name(self):
        return self._story['name']

    def get_url(self):
        return self._story['url']

    def get_type(self):
        return self._story['story_type']

    def get_state(self):
        return self._story['current_state']

    def set_state(self, state):
        self._update(current_state=state)

    def get_labels(self):
        return self._story.get('labels', [])

    def add_label(self, label):
        labels = self.get_labels()
        if label in labels:
            return
        labels.append(label)
        self._update(labels=labels)

    def is_labeled(self, label):
        return label in self.get_labels()

    def add_comment(self, comment):
        self._client.stories.add_comment(
            project_id=self._project_id, story_id=self.get_id(), text=comment)

    def start(self):
        self.set_state('started')

    def is_started(self):
        return self.get_state() == 'started'

    def is_unstarted(self):
        return self.get_state() == 'unstarted'

    #+++ Bug- & Feature-specific stuff
    def finish(self):
        assert self.is_feature() or self.is_bug()
        if self.get_state() != 'started':
            raise StatusError('Feature not started: ' + str(self.get_id()))
        self.set_state('finished')

    def is_finished(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'finished'

    def deliver(self):
        assert self.is_feature() or self.is_bug()
        if self.get_state() != 'finished':
            raise StatusError('Feature not finished: ' + str(self.get_id()))
        self.set_state('delivered')

    def is_delivered(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'delivered'

    def is_rejected(self):
        assert self.is_feature() or self.is_bug()
        return self.get_state() == 'rejected'

    def get_release(self):
        assert self.is_feature() or self.is_bug()
        for label in self.get_labels():
            m = re.match('release-' + _get_version_matcher() + '$', label)
            if m:
                return m.groups()[0]

    def assign_to_release(self, release):
        assert self.is_feature() or self.is_bug()
        _check_version_format(release)
        if self.get_release():
            raise ReleaseAlreadyAssigned('Story already assigned to a release')
        self.add_label('release-' + release)

    def get_branch(self):
        assert self.is_feature() or self.is_bug()
        mgr = _gitflow.managers['feature']
        try:
            return mgr.by_name_prefix(str(self.get_id()))
        except NoSuchBranchError:
            pass
        try:
            return mgr.by_name_prefix(str(self.get_id()), remote=True)
        except NoSuchBranchError:
            pass
        return None
    #--- Bug- & Feature-specific stuff

    #+++ Feature-specific stuff
    def is_feature(self):
        return self.get_type() == 'feature'

    def get_estimate(self):
        assert self.is_feature()
        return self._story['estimate']

    def is_estimated(self):
        assert self.is_feature()
        return self.get_estimate() != -1
    #--- Feature-specific stuff

    #+++ Bug-specific stuff
    def is_bug(self):
        return self.get_type() == 'bug'
    #--- Bug-specific stuff

    #+++ Chore-specific stuff
    def is_chore(self):
        return self.get_type() == 'chore'
    #--- Chore-specific stuff

    def dump(self, highlight_labels=[]):
        story = self._story
        sys.stdout.write(colorize_string(story['name']))
        labels = []
        for label in story.get('labels', []):
            if label in highlight_labels:
                lblstr = Style.BRIGHT + str(label) + Style.RESET_ALL
            else:
                lblstr = str(label)
            labels.append(lblstr)
        sys.stdout.write(' ' + Style.DIM + ','.join(labels) + Style.RESET_ALL)
        sys.stdout.write(' (' + Style.DIM + str(story['current_state']) +
            Style.RESET_ALL + ')')
        sys.stdout.write('\n')

    def _update(self, **kwargs):
        try:
            self._client.stories.update(project_id=self._project_id,
                                        story_id=self.get_id(),
                                        **kwargs)
        except pt.RequestError, e:
            msg  = e.parsed_body['message']
            msg += '\n\nMake sure that you are allowed to update this story!'
            raise GitflowError(msg)
        # Commit changes into our internal story instance as well.
        self._story.update(kwargs)

    def to_dict(self):
        return self._story

    @classmethod
    def from_dict(cls, story):
        t = type('Story', (cls,), dict(_story=story))
        return t(0, _skip_story_download=True)


class Release(object):
    def __init__(self, version, _skip_story_download=False):
        _check_version_format(version)
        self._version = version
        if _skip_story_download:
            return
        self._current_stories = list(_iter_current_stories())

    def start(self):
        for story in self.iter_candidates():
            story.assign_to_release(self._version)

    def try_deliver(self):
        err = False
        for story in self.iter_stories():
            if not story.is_labeled('qa+'):
                err = True
                print "    Story not QA'd: " + story.get_url()
        if err:
            raise StatusError("There were some stories that were not QA's, operation canceled.")

    def deliver(self):
        print 'Following stories were delivered as of release %s:' \
              % self._version
        for story in self.iter_stories():
            story.deliver()
            sys.stdout.write('    ')
            story.dump()

    def prompt_for_confirmation(self):
        answer = raw_input("Do you wish to start the release? [y/N]: ")
        if answer.lower() == 'y':
            return True
        return False

    def iter_stories(self):
        for story in self._current_stories:
            if story.is_labeled('release-' + self._version):
                yield story

    def iter_candidates(self):
        for story in self._current_stories:
            if (story.is_feature() or story.is_bug()) \
                    and story.is_finished() \
                    and story.get_release() is None:
                yield story

    def dump_stories(self):
        print '%s %s %s %s' % (32 * '-', 'Release', self._version, 33 * '-')
        for story in self.iter_stories():
            story.dump(highlight_labels=['release-' + self._version])
        print 80 * '-' + '\n'

    @classmethod
    def dump_all_releases(cls):
        stories = dict()
        for story in _iter_current_stories():
            if story.is_feature() or story.is_bug():
                release = story.get_release()
                if release:
                    if release in stories:
                        stories[release].append(story)
                    else:
                        stories[release] = [story]
        for release in stories:
            cls.from_dict(release, stories[release]).dump_stories()

    @classmethod
    def from_dict(cls, version, stories):
        t = type('Release', (cls,), dict(_current_stories=stories))
        return t(version, _skip_story_download=True)


def print_story(story, index=None, highlight_labels=[]):
    if index:
        sys.stdout.write(Style.BRIGHT + str(index) + ' : ' + Style.RESET_ALL)
    sys.stdout.write(colorize_string(story['name']))
    labels = []
    for label in story.get('labels', []):
        if label in highlight_labels:
            lblstr = Style.BRIGHT + str(label) + Style.RESET_ALL
        else:
            lblstr = str(label)
        labels.append(lblstr)
    sys.stdout.write(' ' + Style.DIM + ','.join(labels) + Style.RESET_ALL)
    sys.stdout.write(' (' + Style.DIM + str(story['current_state']) +
        Style.RESET_ALL + ')')
    sys.stdout.write('\n')


def filter_stories(stories, states, types=None):
    types = types or ['feature', 'chore', 'bug']
    return [s for s in stories if s['current_state'] in states and
        s['story_type'] in types]


def prompt_user_to_select_story():
    i = 1
    stories = list()
    print Style.DIM + "--------- Stories -----------" + Style.RESET_ALL
    for story in _iter_stories():
        if story.is_chore():
            continue
        if story.is_feature() and not story.is_estimated():
            continue
        # You do not start a story if its branch is present or it is
        # assigned to a release, unless it is rejected.
        # If it is not rejected, use checkout instead.
        if (story.get_release() is not None \
                or story.get_branch() is not None) \
                and not story.is_rejected():
            continue
        if story.is_started() \
                or story.is_unstarted() \
                or story.is_rejected():
            stories.append(story)
            print_story(story.to_dict(), i)
            i += 1
    print Style.DIM + "-----------------------------" + Style.RESET_ALL
    print

    if len(stories) == 0:
        raise SystemExit('No story available, aborting!\n\n' \
                'Try "git flow feature list" to find another story' \
                ' to work on.')

    # Prompt user to choose story index.
    print "Select story (or 'q' to quit): "
    index = raw_input()
    if index == "q":
        raise SystemExit("Operation canceled.")

    # We expect a number.
    try:
        index = int(index)
    except ValueError:
        raise GitflowError("Expected a number.")

    if not 0 < index <= len(stories):
        raise GitflowError("Invalid story index: %s." % index)

    # We're expecting input to start from 1, so we have to
    # subtract one here to get the list index.
    story = stories[index - 1]
    slug = prompt_user_to_select_slug(story)

    return [story.get_id(), str(story.get_id()) + '/' + slug]


def prompt_user_to_select_slug(story):
    print ("Please choose short story description (will be slugified " +
        "and used as part of the branch name). Don't use whitespace.")
    slug_hint = slugify(get_story_bold_part(story.get_name()))
    print "Story description (default: '%s'):" % slug_hint
    slug = raw_input().strip() or slug_hint
    if any(c in (string.whitespace + '/') for c in slug):
        print ("\n%sERROR%s: The slug must not contain whitespace or '/'.\n" %
            (Style.BRIGHT, Style.RESET_ALL))
        return prompt_user_to_select_slug(story)
    return slug


def get_story_bold_part(story_name):
    """
        >>> get_story_bold_part("*This is the bold part*, this is not.")
        'This is the bold part'
    """
    if not '*' in story_name:
        return story_name
    else:
        return re.sub(r'.*\*(.+)\*.*', r'\1', story_name)


def slugify(story_name, max_length=25):
    """
        >>> slugify(' Just a story name ')
        'just-a-story-name'
        >>> slugify('Just_a_story_name')
        'just_a_story_name'
        >>> slugify('Just_a_story_name_but_a_very_long_one_indeed...')
        'just_a_story_name_but_a_'
        >>> slugify('Just a story name but a very long one indeed...')
        'just-a-story-name-but-a'
    """
    # Inspired by Django.
    value = re.sub('[^\w\s-]', '', story_name).strip().lower()
    value = re.sub('[-\s]+', '-', value)
    if len(value) > max_length:
        value = value[:max_length]
        return value[:(value.rfind('-'))]
    else:
        return value


def get_iterations():
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    current = client.iterations.current(project_id)
    backlog = client.iterations.backlog(project_id)
    return [current, backlog]


def update_story(story_id, **kwargs):
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    try:
        client.stories.update(
            project_id=project_id, story_id=story_id, **kwargs)
    except pt.RequestError, e:
        msg  = e.parsed_body['message']
        msg += '\n\nMake sure that you are allowed to update this story!'
        raise GitflowError(msg)


def add_comment_to_story(story_id, msg):
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    client.stories.add_comment(
        project_id=project_id, story_id=story_id, text =msg)


def get_story(story_id):
    token = _gitflow.get('gitflow.pt.token')
    project_id = _gitflow.get('gitflow.pt.projectid')
    client = pt.PivotalClient(token=token)
    return client.stories.get(project_id=project_id, story_id=story_id)


def finish_story(story_id):
    story = get_story(story_id)
    update_story(story_id, current_state='finished')


def colorize_string(string):
    """ Use ANSI color codes to emulate a simple subset of textile formatting.
        We're supporting '*' (bold).

        Note: Requires 'termcolor' Python module. If not installed, this
        function just returns the string passed in.
    """
    # Remove escaped parts.
    tmp = re.sub('==.*==', '', string)
    bolds = re.findall('\*.+\*', tmp)
    for bold in bolds:
        stripped_asterisks = (Style.BRIGHT + bold.replace('*', '') +
            Style.RESET_ALL)
        tmp = tmp.replace(bold, stripped_asterisks)
    return tmp

def get_story_id_from_branch_name(branch_name):
    match = re.match('^.+/([0-9]+)[/-]?.*$', branch_name)
    if not match:
        # Gitflow identifier has already been stripped.
        match = re.match('^([0-9]+)[/-]?.*$', branch_name)
        if not match:
            raise GitflowError('Weird branch name format: %s' % branch_name)
    return match.groups()[0]


def show_release_summary(gitflow):
    current = get_iterations()[0]
    current_stories = [
        story for i in current['iterations'] for story in i['stories']]

    print Style.BRIGHT + 'Releasable stories:' + Style.RESET_ALL
    candidate_chores = filter_stories(
        current_stories, ['started'], types=['chore'])
    releasable_chores = [chore for chore in candidate_chores
        if 'reviewed' in chore.get('labels', [])]
    candidate_stories = filter_stories(
        current_stories, ['finished'], types=['bug', 'feature'])
    releasable_stories = [s for s in candidate_stories
        if set(['reviewed', 'qa+']).issubset(set(s.get('labels', [])))]
    for item in releasable_stories + releasable_chores:
        sys.stdout.write(' > ')
        print_story(item)

    print Style.BRIGHT + 'Unreleasable stories:' + Style.RESET_ALL
    unreleasable = [item for item in current_stories
        if not item in releasable_chores and not item in releasable_stories]
    for item in unreleasable:
        sys.stdout.write(' < ')
        print_story(item, highlight_labels=['qa+', 'reviewed'])

    #TODO(Tom): Better release summary based on what's already been merged...
    #lines = gitflow.git.log(
    #    '--show-notes=workflow', '--oneline', '--format="%H %N"').split('\n')
    #finished_stories = []
    #for line in lines:
    #    parts = line.split(' ')
    #    if len(parts) == 3:
    #        (hash, op, story_id) = parts
    #        if op == 'finish':
    #            finished_stories.append((hash, story_id))

    #lines = gitflow.git.log('--oneline', '--format="%h %s"').split('\n')
    ##merges = [l for l in lines if l.split(' ')[1].lower() == 'merge']
    #merged_branches = []
    #for line in lines:
    #    match = re.match(
    #        "(.+)\ Merge\ branch\ '(.+)'\ into %s" % gitflow.develop_name(),
    #        line)
    #    if match:
    #        (hash, branch) = match.groups()
    #        if not branch in merged_branches:
    #            merged_branches.append(branch)
    #print merged_branches
    #releasable = releasable_stories + releasable_chores
    #for story in releasable:
    #    stories_merged_before_this_story = \
    #        set([get_story_id_from_branch_name(s) for s in \
    #            merged_branches[merged_branches.index(story) + 1 : ]])
    #    blockers = stories_merged_before_this_story.intersection(
    #        [s['id'] for s in unreleasable])
    #    if blockers:
    #        print "Story '%s' is blocked by stories: %s" % (story, blockers)
    #import ipdb; ipdb.set_tracce()



#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

class GitflowError(Exception): pass

class Usage(GitflowError):
    def __str__(self):
        return '\n'.join(map(str, self.args))

class OperationsError(GitflowError): pass
class PrefixNotUniqueError(OperationsError): pass
class MergeError(OperationsError): pass


class StatusError(GitflowError): pass
class WorkdirIsDirtyError(StatusError): pass
class NotInitialized(StatusError): pass
class AlreadyInitialized(StatusError):
    def __str__(self):
        return ("Already initialized for gitflow.\n"
                "To force reinitialization use: git flow init -f")
class ReleaseAlreadyAssigned(StatusError): pass
class MultipleReviewRequestsForBranch(StatusError):
    def __str__(self):
        return 'Multiple review requests for {0} were found'.format(self.args[0])
class IllegalCommunicationProtocol(StatusError): pass
class RemoteNotFound(StatusError): pass

class MergeConflict(StatusError):
    def __str__(self):
        return '\n'.join([
            "Merge conflicts not resolved yet, use:",
            "    git mergetool",
            "    git commit",
            ])


class ObjectError(GitflowError): pass
class BadObjectError(ObjectError): pass
class NoSuchBranchError(ObjectError): pass
class NoSuchRemoteError(ObjectError): pass

class BaseNotOnBranch(ObjectError):
    def __str__(self):
        return ("Branch '%s' found in the repository," \
                "but it is diverging from '%s'" \
                % (self.args[1], self.args[0]))

class BaseNotFound(ObjectError): pass
class BaseNotAllowed(ObjectError): pass

class BranchExistsError(ObjectError):pass
class TagExistsError(ObjectError): pass

class BranchTypeExistsError(ObjectError):
    def __str__(self):
        return("There is an existing %s branch. "
                "Finish that one first." % self.args[0])

class NoSuchLocalBranchError(NoSuchBranchError):
    def __str__(self):
        return ("Local branch '%s' does not exist." % self.args[0])


class IllegalVersionFormat(GitflowError):
    def __str__(self):
        return "Version string is ill-formed, it is not matching " + \
                self.args[0]



"""
git-flow -- A collection of Git extensions to provide high-level
repository operations for Vincent Driessen's branching model.
"""
#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

VERSION = (0, 7, 0, 'rc13')

__version__ = ".".join(map(str, VERSION[0:3])) + "".join(VERSION[3:])
__author__ = "Ondrej Kupka, Tomas Brambora, Hartmut Goebelm, Vincent Driessen"
__contact__ = "ondrejk@salsitasoft.com, tomas@salsitasoft.com, h.goebel@crazy-compilers.com, vincent@datafox.nl"
__homepage__ = "http://github.com/salsita/gitflow/"
__docformat__ = "restructuredtext"
__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel; 2013 Tomas Brambora, 2013 Ondrej Kupka"
__license__ = "BSD"



#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from git import GitCommandError, Reference
from gitflow.exceptions import (NoSuchBranchError, BranchExistsError,
                                PrefixNotUniqueError, BaseNotOnBranch,
                                WorkdirIsDirtyError, BranchTypeExistsError,
                                TagExistsError, MergeError, BaseNotFound)

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

class BranchManager(object):
    """
    Initializes an instance of :class:`BranchManager`.  A branch
    manager is responsible for listing, creating, merging, deleting,
    finishing (i.e. merging+deleting) branches of a given type.

    :param gitflow:
        The :class:`gitflow.core.GitFlow` instance that this branch
        manager belongs to.

    :param prefix:
        The prefix to use for the type of branches that this branch
        manager manages. If this is not `None`, it supersedes the
        configuration of the `gitflow` object's repository.
    """

    def _get_prefix(self):
        if self._prefix is not None:
            return self._prefix
        try:
            return self.gitflow.get_prefix(self.identifier)
        except:
            return self.DEFAULT_PREFIX
    def _set_prefix(self, value): self._prefix = value
    prefix = property(_get_prefix, _set_prefix)

    def __init__(self, gitflow, prefix=None):
        from gitflow.core import GitFlow
        assert isinstance(gitflow, GitFlow), "Argument 'gitflow' must be a GitFlow instance."
        self.gitflow = gitflow
        if not prefix is None:
            assert isinstance(prefix, basestring), "Argument 'prefix' must be a string."
            self.prefix = prefix
        else:
            self.prefix = None

    def default_base(self):
        """
        :returns:
            The name of branch to use as the default base for branching off from
            in case no explicit base is specified.

        This method can be overriden in a subclass of :class:`BranchManager`.
        If not overriden, the default is to use the 'develop' branch.
        """
        return self.gitflow.develop_name()

    def full_name(self, name):
        return self.prefix + name

    def shorten(self, full_name):
        """
        Returns the friendly (short) name of this branch, without the prefix,
        given the fully qualified branch name.

        :param full_name:
            The full name of the branch as it is known to Git, including the
            prefix.

        :returns:
            The friendly name of the branch.
        """
        if full_name.startswith(self.prefix):
            return full_name[len(self.prefix):]
        else:
            return full_name

    def by_name_prefix(self, nameprefix, remote=False):
        """
        If exactly one branch of the type that this manager manages starts with
        the given name prefix, returns that branch.  Raises
        :exc:`NoSuchBranchError` in case no branches exist with the given
        prefix, or :exc:`PrefixNotUniqueError` in case multiple matches are
        found.

        :param nameprefix:
            The name prefix (or full name) of the short branch name to match.

        :returns:
            The :class:`git.refs.Head` instance of the branch that can be
            uniquely identified by the given name prefix.
        """
        if remote:
            nameprefix = self.gitflow.origin_name(self.full_name(nameprefix))
        else:
            nameprefix = self.full_name(nameprefix)
        matches = [b
                   for b in self.iter(remote)
                   if b.name.startswith(nameprefix)]
        num_matches = len(matches)
        if num_matches == 1:
            return matches[0]
        elif num_matches < 1:
            raise NoSuchBranchError('There is no %s branch matching the '
                    'prefix "%s"' % (self.identifier, nameprefix))
        else:
            raise PrefixNotUniqueError('There are multiple %s branches '
                    'matching the prefix "%s": %s' % (self.identifier,
                        nameprefix, matches))

    def iter(self, remote=False):
        """
        :returns:
            An iterator, iterating over all branches of the type that this
            manager manages.
        """
        if remote:
            refs = self.gitflow.repo.refs
            prefix = self.gitflow.origin_name(self.prefix)
        else:
            refs = self.gitflow.repo.branches
            prefix = self.prefix

        for r in refs:
            if r.name.startswith(prefix):
                yield r

    def list(self, remote=False):
        """
        :returns:
            A list of all branches of the type that this manager manages.  See
            also :meth:`iter`.
        """
        return list(self.iter(remote))

    def create(self, name, base=None, fetch=False,
               must_be_on_default_base=False):
        """
        Creates a branch of the type that this manager manages and checks it
        out.

        :param name:
            The (short) name of the branch to create.

        :param base:
            The base commit or ref to base the branch off from.  If a base is
            not provided explicitly, the default base for this type of branch is
            used.  See also :meth:`default_base`.

        :param fetch:
            If set, update the local repo with remote changes prior to
            creating the new branch.

        :param must_be_on_default_base:
            If set, the `base` must be a valid commit on the branch
            manager `default_base`.

        :returns:
            The newly created :class:`git.refs.Head` reference.
        """
        gitflow = self.gitflow
        repo = gitflow.repo

        if fetch:
            gitflow.origin().fetch()

        # Make sure the branch exists neither locally nor remotely.
        full_name = self.full_name(name)
        origin_name = gitflow.origin_name(full_name)
        for r in repo.refs:
            if r.name in (full_name, origin_name):
                raise BranchExistsError(full_name)

        # Make sure the base exists and is in sync.
        if base is None:
            base = self.default_base()
        elif must_be_on_default_base:
            if not gitflow.is_merged_into(base, self.default_base()):
                raise BaseNotOnBranch(base, self.default_base())

        origin_base = gitflow.origin_name(base)
        if base in repo.refs:
            if origin_base in repo.refs:
                gitflow.require_branches_equal(origin_base, base)
        elif origin_base in repo.refs:
            base = origin_base
        else:
            raise BaseNotFound('Base branch {0} not found.'.format(base))

        # Make sure there are no merge conflicts.
        gitflow.require_no_merge_conflict()
        if gitflow.has_staged_commits():
            raise WorkdirIsDirtyError('Contains local changes checked into '
                                      'the index but not committed.')

        # Finally, create and checkout.
        branch = repo.create_head(full_name, base)
        branch.checkout()
        return branch

    def _is_single_commit_branch(self, from_, to):
        git = self.gitflow.repo.git
        commits = git.rev_list('%s...%s' % (from_, to), n=2).split()
        return len(commits) == 1

    def merge(self, name, into, message=None):
        """
        This merges the branch named :attr:`name` into the branch named
        :attr:`into`, using commit message :attr:`message`.

        :param name:
            The (short) name of the branch that needs merging.  Alternatively
            this is a instance of git.Reference (or subclass).
        :param into:
            The name of the branch to merge into.
        :param message:
            The commit message to use for the merge commit.  If it is not given,
            a default merge message is used.  You can use the following string
            placeholders, which :meth:`merge` will expand::

                %(name)s = The full name of the branch (e.g. 'feature/foo')
                %(short_name)s = The friendly name of the branch (e.g. 'foo')
                %(identifier)s = The type (e.g. 'feature', 'hotfix', etc.)

        You typically don't need to override this method in a subclass.
        """
        if isinstance(name, basestring):
            full_name = self.prefix + name
        else:
            assert isinstance(name, Reference)
            full_name = name
        repo = self.gitflow.repo
        repo.branches[into].checkout()
        if self.gitflow.is_merged_into(full_name, into):
            # already merged, nothing more to do
            return

        kwargs = dict(no_ff=True)
        if message is not None:
            message = (message
                       % dict(name=full_name, identifier=self.identifier,
                              short_name=name))
            kwargs['message'] = message
        # `git merge` does not send the error message to stderr, thus
        # we need to capture stdout manually :-(
        stdout = StringIO()
        try:
            repo.git.merge(full_name, output_stream=stdout, **kwargs)
        except GitCommandError, e:
            txt = stdout.getvalue().rstrip()
            if e.stderr:
                text = text + '\n' + e.stderr
            raise MergeError(txt)

    def delete(self, name, force=False):
        """
        This deletes a branch of the type that this manager manages named
        :attr:`name`.

        :param name:
            The (short) name of the branch to delete.

        :param force:
            Delete the branch, even if this would lead to data loss.
        """
        repo = self.gitflow.repo
        full_name = self.prefix + name
        repo.delete_head(full_name, force=force)


class FeatureBranchManager(BranchManager):
    identifier = 'feature'
    DEFAULT_PREFIX = 'feature/'

    def create(self, name, base=None, fetch=False):
        """
        Creates a branch of type `feature` and checks it out.

        :param name:
            The (short) name of the branch to create. This will be
            prefixed by `feature/` or whatever is configured in
            `gitflow.prefix.feature`.

        :param base:
            The base commit or ref to base the branch off from.  If no
            base is provided, it defaults to the branch configured in
            `gitflow.branch.develop`.  See also :meth:`default_base`.

        :param fetch:
            If set, update the local repo with remote changes prior to
            creating the new branch.

        :returns:
            The newly created :class:`git.refs.Head` reference.
        """
        # :todo:rethink: feature branch is not required to start at `develop`
        return super(FeatureBranchManager, self).create(
            name, base, fetch=fetch, must_be_on_default_base=False)


    def finish(self, name, fetch=False, upstream='default', rebase=False,
               keep=False, force_delete=False, push=False, tagging_info=None):
        """
        Finishes the branch of type `feature` named :attr:`name`.
        Finishing means that:

        * The `feature`-branch is merged into the `develop`-branch.
        * The `feature`-branch is deleted if all merges are successful.

        :param name:
            The (short) name of the branch to finish.
        :param fetch:
            If set, update the local repo with remote changes prior to
            merging.
        :param rebase:
            Rebase `featuer`-branch prior to merging.
        :param keep:
            Keep `feature`-branch after performing finish.
        :param force_delete:
            Force deleting the `feature`-branch even if merging failed.
        :param push:
            Push changes to the remote repository.
        """
        assert not tagging_info, "FeatureBranchManager does not support tagging"
        gitflow = self.gitflow
        full_name = self.full_name(name)
        gitflow.must_be_uptodate(full_name, fetch=fetch)
        gitflow.must_be_uptodate(upstream, fetch=fetch)

        if rebase:
            gitflow.rebase(self.identifier, name, interactive=False,
                           upstream=upstream)
        else:
            self.merge(name, upstream,
                       'Finished %(identifier)s %(short_name)s.')

        to_push = [upstream]

        if not keep:
            self.delete(name, force=force_delete)
            to_push.append(':'+full_name)
        if push:
            gitflow.origin().push(to_push)


class ReleaseBranchManager(BranchManager):
    identifier = 'release'
    DEFAULT_PREFIX = 'release/'

    def create(self, version, base=None, fetch=False):
        """
        Creates a branch of type `release` and checks it out.

        :param version:
            The version to be released and for which to create the
            release-branch for. This will be prefixed by `release/`
            or whatever is configured in `gitflow.prefix.release`.

        :param base:
            The base commit or ref to base the branch off from.  If no
            base is provided, it defaults to the branch configured in
            `gitflow.branch.develop`.  See also :meth:`default_base`.

        :param fetch:
            If set, update the local repo with remote changes prior to
            creating the new branch.

        :returns:
            The newly created :class:`git.refs.Head` reference.
        """
        gitflow = self.gitflow
        git = gitflow.git

        # there must be no active `release` branch
        if len(self.list()) > 0:
            raise BranchTypeExistsError(self.identifier)

        # there must be no tag for this version yet
        tagname = gitflow.get('gitflow.prefix.versiontag') + version
        if tagname in gitflow.repo.tags:
            raise TagExistsError(tagname)

        # create the release branch
        b = super(ReleaseBranchManager, self).create(
            version, base, fetch=fetch, must_be_on_default_base=True)

        # push the newly created branch
        git.push(gitflow.origin_name(), self.by_name_prefix(version))

        return b


    def finish(self, name, fetch=False, rebase=False, keep=False,
               force_delete=False, push=False, tagging_info=None):
        assert rebase == False, "Rebasing a release branch does not make any sense."
        # require release branch to exist
        # if flag-fetch: fetch master und develop
        #   diese muessen dann gleich $ORIGIN/master bzw. $ORIGIN/develop sein
        gitflow = self.gitflow
        full_name = self.full_name(name)
        gitflow.must_be_uptodate(full_name, fetch=fetch)
        gitflow.must_be_uptodate(gitflow.develop_name(), fetch=fetch)
        gitflow.must_be_uptodate(gitflow.master_name(), fetch=fetch)

        to_push = [self.gitflow.develop_name(), self.gitflow.master_name()]

        self.merge(name, self.gitflow.master_name(),
                'Finished %s %s.' % (self.identifier, name))

        tag = None
        if tagging_info is not None:
            # try to tag the release
            tagname = self.gitflow.get('gitflow.prefix.versiontag') + name
            # In case a previous attempt to finish this release branch
            # has failed, but the tag was set successful, we skip it
            # now.
            # :todo: check: if tag exists, it must point to the commit
            tag = gitflow.tag(tagname, self.gitflow.master_name(),
                        **tagging_info)
            to_push.append(tagname)

        # merge the master branch back into develop; this makes the
        # master branch - and the new tag (if provided) - a parent of
        # the development branch, which in turn lets you use 'git
        # describe' on either branch
        self.merge(tag or self.gitflow.master(),
                   self.gitflow.develop_name(),
                   'Finished %s %s.' % (self.identifier, name))
        if not keep:
            self.delete(name, force=force_delete)
            to_push.append(':'+full_name)
        if push:
            gitflow.origin().push(to_push)


class HotfixBranchManager(ReleaseBranchManager):
    identifier = 'hotfix'
    DEFAULT_PREFIX = 'hotfix/'

    def default_base(self):
        return self.gitflow.master_name()


class SupportBranchManager(BranchManager):
    identifier = 'support'
    DEFAULT_PREFIX = 'support/'

    def default_base(self):
        return self.gitflow.master_name()

    def create(self, name, base=None, fetch=False):
        """
        Creates a branch of type `support` and checks it out.

        :param name:
            The (short) name of the branch to create. This will be
            prefixed by `support/` or whatever is configured in
            `gitflow.prefix.support`.

        :param base:
            The base commit or ref to base the branch off from.  If no
            base is provided, it defaults to the branch configured in
            `gitflow.branch.develop`.  See also :meth:`default_base`.

        :param fetch:
            If set, update the local repo with remote changes prior to
            creating the new branch.

        :returns:
            The newly created :class:`git.refs.Head` reference.
        """
        return super(SupportBranchManager, self).create(
            name, base, fetch=fetch, must_be_on_default_base=True)

    def finish(self, *args):
        raise NotImplementedError("Finishing support branches does not make "
                "any sense.")



#
# This file is part of `gitflow`.
# Copyright (c) 2010-2011 Vincent Driessen
# Copyright (c) 2012 Hartmut Goebel
# Distributed under a BSD-like license. For full terms see the file LICENSE.txt
#

from unittest2 import TestCase
import os
import sys
try:
    import cStringIO as StringIO
except:
    import StringIO
from ConfigParser import NoOptionError, NoSectionError

from git import GitCommandError

from gitflow.core import GitFlow
from gitflow.branches import BranchManager
from gitflow.exceptions import (BranchExistsError, NotInitialized, MergeError,
                                NoSuchBranchError, NoSuchRemoteError,
                                MergeConflict, BadObjectError, Usage)
from tests.helpers import (copy_from_fixture, remote_clone_from_fixture,
                           fake_commit, all_commits, set_gnupg_home)
from tests.helpers.factory import create_sandbox, create_git_repo

__copyright__ = "2010-2011 Vincent Driessen; 2012 Hartmut Goebel"
__license__ = "BSD"

class TestGitFlowBasics(TestCase):

    # Configuration
    @copy_from_fixture('custom_repo')
    def test_config_reader(self):
        gitflow = GitFlow(self.repo).init()
        self.assertRaises(ValueError, gitflow.get,
                'invalid_setting_since_this_has_no_dot')
        self.assertRaises(NoSectionError, gitflow.get,
                'nonexisting.nonexisting')
        self.assertRaises(NoSectionError, gitflow.get,
                'section.subsection.propname')
        self.assertRaises(NoOptionError, gitflow.get, 'foo.nonexisting')
        self.assertEquals('qux', gitflow.get('foo.bar'))

    @copy_from_fixture('custom_repo')
    def test_custom_branchnames(self):
        gitflow = GitFlow(self.repo).init()
        self.assertEquals('production', gitflow.master_name())
        self.assertEquals('master', gitflow.develop_name())
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('f-', gitflow.get_prefix('feature'))
        self.assertEquals('rel-', gitflow.get_prefix('release'))
        self.assertEquals('hf-', gitflow.get_prefix('hotfix'))
        self.assertEquals('supp-', gitflow.get_prefix('support'))
        self.assertEquals('v', gitflow.get_prefix('versiontag'))


    # Initialization
    def test_branch_names_fails_in_new_sandbox(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox)
        self.assertRaises(NotInitialized, gitflow.branch_names)

    def test_empty_repo_has_no_branches(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        self.assertItemsEqual([], gitflow.branch_names())

    @copy_from_fixture('custom_repo')
    def test_custom_repo_has_branches(self):
        gitflow = GitFlow(self.repo).init()
        self.assertItemsEqual(['master', 'production'],
                gitflow.branch_names())

    @copy_from_fixture('custom_repo')
    def test_custom_repo_init_keeps_active_branch_if_develop_already_existed(self):
        active_branch = self.repo.active_branch
        gitflow = GitFlow(self.repo).init()
        self.assertNotEqual(gitflow.repo.active_branch.name, active_branch)

    def test_origin_without_remote_raises_error(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        self.assertRaises(NoSuchRemoteError, gitflow.origin)

    # Sanity checking
    def test_new_repo_is_not_dirty(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox).init()
        self.assertFalse(gitflow.is_dirty())

    @copy_from_fixture('dirty_sample_repo')
    def test_existing_repo_is_not_dirty(self):
        gitflow = GitFlow(self.repo)
        self.assertTrue(gitflow.is_dirty())

    def test_new_repo_has_no_staged_commits(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox).init()
        self.assertFalse(gitflow.has_staged_commits())

    @copy_from_fixture('dirty_sample_repo')
    def test_existing_repo_has_staged_commits(self):
        gitflow = GitFlow(self.repo)
        self.assertTrue(gitflow.has_staged_commits())

    def test_gitflow_cannot_get_status_on_empty_sandbox(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox)
        self.assertRaises(NotInitialized, gitflow.status)

    def test_gitflow_status_on_fresh_repo(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        self.assertEquals([], gitflow.status())

    @copy_from_fixture('sample_repo')
    def test_gitflow_status_on_sample_repo(self):
        gitflow = GitFlow(self.repo)
        self.assertItemsEqual([
                ('stable', '296586bb164c946cad10d37e82570f60e6348df9', False),
                ('devel', '2b34cd2e1617e5f0d4e077c6ec092b9f50ed49a3', False),
                ('feat/recursion', '54d59c872469c7bf34d540d2fb3128a97502b73f', True),
                ('feat/even', 'e56be18dada9e81ca7969760ddea357b0c4c9412', False),
            ], gitflow.status())


    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_status_on_cloned_sample_repo(self):
        gitflow = GitFlow(self.repo)
        # NB: only the active branch is created locally when cloning
        self.assertEqual(self.remote.active_branch.name, 'feat/recursion')
        self.assertItemsEqual([
                ('feat/recursion', '54d59c872469c7bf34d540d2fb3128a97502b73f', True),
            ], gitflow.status())

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_status_on_remote_sample_repo(self):
        gitflow = GitFlow(self.remote)
        self.assertItemsEqual([
                ('stable', '296586bb164c946cad10d37e82570f60e6348df9', False),
                ('devel', '2b34cd2e1617e5f0d4e077c6ec092b9f50ed49a3', False),
                ('feat/recursion', '54d59c872469c7bf34d540d2fb3128a97502b73f', True),
                ('feat/even', 'e56be18dada9e81ca7969760ddea357b0c4c9412', False),
            ], gitflow.status())


    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_cloned_sample_repo_is_not_remote(self):
        self.assertNotEqual(self.remote.git_dir, self.repo.git_dir)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_cloned_sample_repo_equals_remote(self):
        # only the active branch (feat/recursion) is cloned which
        # includes devel and stable
        heads_to_check = [h.name for h in self.remote.heads]
        heads_to_check.remove('feat/even')
        self.assertIn('devel', heads_to_check)
        self.assertIn('stable', heads_to_check)
        all_remote_commits = all_commits(self.remote, heads_to_check)
        all_cloned_commits = all_commits(self.repo)
        self.assertEqual(all_remote_commits, all_cloned_commits)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_cloned_sample_repo_remote_name(self):
        self.assertEqual(len(list(self.repo.remotes)), 1)
        self.assertEqual(self.repo.remotes[0].name, 'my-remote')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_required_remote_returns_remote(self):
        gitflow = GitFlow(self.repo).init()
        remote = gitflow.require_remote('my-remote')
        self.assertEqual(self.repo.remotes['my-remote'], remote)

    def test_gitflow_required_remote_raises_error_on_nonexisting_remote(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox).init()
        self.assertRaises(NoSuchRemoteError,
                          gitflow.require_remote, 'some-remote')

class TestGitFlowInit(TestCase):
    # git flow init

    def test_gitflow_init_inits_underlying_git_repo(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox)
        dot_git_dir = os.path.join(sandbox, '.git')
        self.assertFalse(os.path.exists(dot_git_dir))
        gitflow.init()
        self.assertTrue(os.path.exists(dot_git_dir))
        self.assertTrue(gitflow.is_initialized())

    def test_gitflow_init_marks_initialized(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        self.assertFalse(gitflow.is_initialized())
        gitflow.init()
        self.assertTrue(gitflow.is_initialized())

    def test_gitflow_throws_errors_before_init(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        self.assertRaises(NotInitialized, gitflow.master_name)
        self.assertRaises(NotInitialized, gitflow.develop_name)
        self.assertRaises(NotInitialized, gitflow.get_prefix, 'feature')

    def test_gitflow_init_initializes_default_config(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        self.assertEquals('master', gitflow.master_name())
        self.assertEquals('develop', gitflow.develop_name())
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('feature/', gitflow.get_prefix('feature'))
        self.assertEquals('hotfix/', gitflow.get_prefix('hotfix'))
        self.assertEquals('release/', gitflow.get_prefix('release'))
        self.assertEquals('support/', gitflow.get_prefix('support'))
        self.assertEquals('', gitflow.get_prefix('versiontag'))

    def test_gitflow_init_with_alternative_config(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo)
        prefixes = dict(feature='f-',
                        hotfix='hf-',
                        release='rel-',
                        support='supp-',
                        versiontag='ver')
        names = dict(origin='somewhereelse')
        gitflow.init(master='foo', develop='bar', prefixes=prefixes, names=names)
        self.assertEquals('foo', gitflow.master_name())
        self.assertEquals('bar', gitflow.develop_name())
        self.assertEquals('somewhereelse', gitflow.origin_name())
        self.assertEquals('f-', gitflow.get_prefix('feature'))
        self.assertEquals('rel-', gitflow.get_prefix('release'))
        self.assertEquals('hf-', gitflow.get_prefix('hotfix'))
        self.assertEquals('supp-', gitflow.get_prefix('support'))
        self.assertEquals('ver', gitflow.get_prefix('versiontag'))

    @copy_from_fixture('partly_inited')
    def test_gitflow_init_config_with_partly_inited(self):
        gitflow = GitFlow(self.repo).init()

        # Already set in fixture, shouldn't change
        self.assertEquals('production', gitflow.master_name())
        self.assertEquals('f-', gitflow.get_prefix('feature'))

        # Implicit defaults
        self.assertEquals('develop', gitflow.develop_name())
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('release/', gitflow.get_prefix('release'))
        self.assertEquals('hotfix/', gitflow.get_prefix('hotfix'))
        self.assertEquals('support/', gitflow.get_prefix('support'))
        self.assertEquals('', gitflow.get_prefix('versiontag'))


    @copy_from_fixture('sample_repo')
    def test_gitflow_init_keeps_active_branch_if_develop_already_existed(self):
        active_branch = self.repo.active_branch.name
        gitflow = GitFlow(self.repo).init()
        self.assertEqual(gitflow.repo.active_branch.name, active_branch)

    def test_gitflow_init_checkout_develop_if_newly_created(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        self.assertEqual(gitflow.repo.active_branch.name, 'develop')

    @copy_from_fixture('sample_repo')
    def test_gitflow_init_creates_no_extra_commits(self):
        all_commits_before_init = all_commits(self.repo)
        gitflow = GitFlow(self.repo).init()
        all_commits_after_init = all_commits(self.repo)
        self.assertEquals(all_commits_before_init, all_commits_after_init)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_init_cloned_creates_no_extra_commits(self):
        all_commits_before_init = all_commits(self.repo)
        gitflow = GitFlow(self.repo).init()
        all_commits_after_init = all_commits(self.repo)
        self.assertEquals(all_commits_before_init, all_commits_after_init)

    @copy_from_fixture('sample_repo')
    def test_gitflow_init_creates_no_extra_branches(self):
        heads_before_init = [h.name for h in self.repo.heads]
        gitflow = GitFlow(self.repo).init()
        heads_after_init = [h.name for h in self.repo.heads]
        self.assertItemsEqual(heads_before_init, heads_after_init)

    def test_gitflow_init_creates_initial_commit(self):
        repo = create_git_repo(self)
        all_commits_before_init = all_commits(repo)
        gitflow = GitFlow(repo).init()
        all_commits_after_init = all_commits(repo)
        self.assertNotEquals(all_commits_before_init, all_commits_after_init)
        self.assertEquals('Initial commit', repo.heads.master.commit.message)

    def test_gitflow_init_creates_branches(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        self.assertItemsEqual(['master', 'develop'],
                gitflow.branch_names())

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_init_cloned_creates_master_and_develop(self):
        heads_before_init = [h.name for h in self.repo.heads]
        self.assertNotIn('stable', heads_before_init)
        self.assertNotIn('devel', heads_before_init)
        gitflow = GitFlow(self.repo).init()
        heads_after_init = [h.name for h in self.repo.heads]
        self.assertIn('stable', heads_after_init)
        self.assertIn('devel', heads_after_init)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_init_cloned_creates_no_custom_master_and_develop(self):
        heads_before_init = [h.name for h in self.repo.heads]
        self.assertNotIn('foo', heads_before_init)
        self.assertNotIn('bar', heads_before_init)
        gitflow = GitFlow(self.repo)
        self.assertRaises(NotImplementedError,
                          gitflow.init, master='foo', develop='bar')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_init_cloned_creates_branches_from_counterpart(self):
        remote =  GitFlow(self.remote)
        rmc0 = remote.master().commit
        rdc0 = remote.develop().commit

        gitflow = GitFlow(self.repo).init()
        mc0 = gitflow.master().commit
        dc0 = gitflow.develop().commit

        # local and remote heads must be the same
        self.assertEqual(rmc0, mc0)
        self.assertEqual(rdc0, dc0)
        self.assertTrue(gitflow.master().tracking_branch())
        self.assertTrue(gitflow.develop().tracking_branch())
        self.assertEqual(gitflow.master().tracking_branch().name, 'my-remote/stable')
        self.assertEqual(gitflow.develop().tracking_branch().name, 'my-remote/devel')


    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_init_cloned_creates_no_extra_banches(self):
        heads_before_init = [h.name for h in self.repo.heads]
        heads_before_init.sort()
        gitflow = GitFlow(self.repo).init()
        heads_after_init = [h.name for h in self.repo.heads]
        heads_after_init.remove('stable')
        heads_after_init.remove('devel')
        self.assertItemsEqual(heads_before_init, heads_after_init)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_init_cloned_checkout_develop_if_newly_created(self):
        gitflow = GitFlow(self.repo).init()
        self.assertEqual(gitflow.repo.active_branch.name, 'devel')

    @copy_from_fixture('partly_inited')
    def test_gitflow_force_reinit_partly_inited(self):
        gitflow = GitFlow(self.repo)
        gitflow.init(force_defaults=True)

        # Implicit defaults
        self.assertEquals('develop', gitflow.develop_name())
        self.assertEquals('origin', gitflow.origin_name())
        self.assertEquals('release/', gitflow.get_prefix('release'))
        self.assertEquals('hotfix/', gitflow.get_prefix('hotfix'))
        self.assertEquals('support/', gitflow.get_prefix('support'))
        self.assertEquals('', gitflow.get_prefix('versiontag'))

        # Explicitly forced back to defaults
        self.assertEquals('master', gitflow.master_name())
        self.assertEquals('feature/', gitflow.get_prefix('feature'))


class TestGitFlowTag(TestCase):

    @copy_from_fixture('sample_repo')
    def test_gitflow_tag_without_message(self):
        gitflow = GitFlow(self.repo).init()
        self.assertNotIn('some-tag', self.repo.tags)
        commit = self.repo.head.commit
        gitflow.tag('some-tag', commit)
        self.assertIn('some-tag', self.repo.tags)
        tagref = self.repo.tags['some-tag']
        self.assertEqual(tagref.commit, commit)
        self.assertEqual(tagref.name, 'some-tag')
        # if there is no message, tagref.tag is None
        self.assertEqual(tagref.tag, None)

    @copy_from_fixture('sample_repo')
    def test_gitflow_tag_with_message(self):
        gitflow = GitFlow(self.repo).init()
        self.assertNotIn('some-tag', self.repo.tags)
        commit = self.repo.head.commit
        gitflow.tag('some-tag', commit, 'This is my tag')
        self.assertIn('some-tag', self.repo.tags)
        tagref = self.repo.tags['some-tag']
        self.assertEqual(tagref.commit, commit)
        self.assertEqual(tagref.name, 'some-tag')
        # if there is no message, tagref.tag is None
        self.assertNotEqual(tagref.tag, None)

        tag = tagref.tag
        self.assertEqual(tag.message, 'This is my tag')
        self.assertEqual(tag.tag, 'some-tag')
        self.assertEqual(tag.object, commit)

    @set_gnupg_home
    @copy_from_fixture('sample_repo')
    def test_gitflow_tag_signed(self):
        gitflow = GitFlow(self.repo).init()
        # need to the the signing key via config
        gitflow.set('user.signingkey', 'Dummy Key for Gitflow testing')
        commit = self.repo.head.commit
        gitflow.tag('some-tag', commit, 'This is my tag', sign=True)
        tag = self.repo.tags['some-tag'].tag
        expected = ['This is my tag', '-----BEGIN PGP SIGNATURE-----']
        self.assertEqual(tag.message.splitlines()[:2], expected)
        self.assertEqual(tag.tag, 'some-tag')
        self.assertEqual(tag.object, commit)

    @set_gnupg_home
    @copy_from_fixture('sample_repo')
    def test_gitflow_tag_signed_with_key(self):
        gitflow = GitFlow(self.repo).init()
        commit = self.repo.head.commit
        gitflow.tag('some-tag', commit, 'This is my tag', sign=True,
                    signingkey='Dummy Key for Gitflow testing')
        tag = self.repo.tags['some-tag'].tag
        expected = ['This is my tag', '-----BEGIN PGP SIGNATURE-----']
        self.assertEqual(tag.message.splitlines()[:2], expected)
        self.assertEqual(tag.tag, 'some-tag')
        self.assertEqual(tag.object, commit)


    # :todo: test-cases for must_be_uptodate
    # :todo: test-cases for branch_names(remote=True)
    # :todo: test-cases for list (one per BranchManager)

class TestGitFlowMerges(TestCase):

    @copy_from_fixture('sample_repo')
    def test_gitflow_is_merged_into(self):
        gitflow = GitFlow(self.repo).init()

        # feat/even is ahead of devel
        self.assertTrue(gitflow.is_merged_into('devel', 'feat/even'))
        self.assertFalse(gitflow.is_merged_into('feat/even', 'devel'))
        # devel as a symbolic ref
        self.assertFalse(gitflow.is_merged_into('feat/even',
                                                self.repo.refs['devel']))
        # feat/even a symbolic ref
        self.assertFalse(gitflow.is_merged_into(self.repo.refs['feat/even'],
                                                'devel'))
        # HEAD
        self.assertFalse(gitflow.is_merged_into('HEAD', 'devel'))
        self.assertFalse(gitflow.is_merged_into(self.repo.head, 'devel'))

    @copy_from_fixture('sample_repo')
    def test_gitflow_is_merged_into_non_existing_source(self):
        gitflow = GitFlow(self.repo).init()
        self.assertRaisesRegexp(BadObjectError, 'feat/ever',
                          gitflow.is_merged_into ,'feat/ever', 'devel')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_is_merged_into_remote(self):
        gitflow = GitFlow(self.repo).init()

        # devel is behind feat/even
        self.assertTrue(gitflow.is_merged_into(
            'devel',
            'remotes/my-remote/feat/even'))
        self.assertTrue(gitflow.is_merged_into(
            'devel',
            gitflow.origin().refs['devel']))
        self.assertTrue(gitflow.is_merged_into(
            gitflow.origin().refs['devel'],
            'remotes/my-remote/feat/even'))
        self.assertTrue(gitflow.is_merged_into(
            gitflow.origin().refs['devel'],
            gitflow.origin().refs['devel']))

        # feat/even is ahead of devel
        self.assertFalse(gitflow.is_merged_into(
            'feat/recursion',
            'devel'))
        self.assertFalse(gitflow.is_merged_into(
            'remotes/my-remote/feat/recursion',
            'devel'))

        self.assertFalse(gitflow.is_merged_into(
            'feat/recursion',
            gitflow.origin().refs['devel']))
        self.assertFalse(gitflow.is_merged_into(
            'remotes/my-remote/feat/recursion',
            gitflow.origin().refs['devel']))

    #:todo: test-case is_merged_into_remote with remote branch beeing ahead
    #       of corresponding local branch

    @copy_from_fixture('sample_repo')
    def test_has_no_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.require_no_merge_conflict()

    @copy_from_fixture('sample_repo')
    def test_has_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        repo = gitflow.repo
        repo.refs['devel'].checkout()
        repo.git.merge('feat/recursion')
        # the next merge creates the merge conflict
        self.assertRaises(GitCommandError,
                          repo.git.merge, 'feat/even')
        self.assertRaises(MergeConflict, gitflow.require_no_merge_conflict)

    # :todo: test-cases for compare_branches
    # :todo: test-cases for require_branches_equal


class TestGitFlowCheckout(TestCase):
    @copy_from_fixture('sample_repo')
    def test_gitflow_nameprefix_or_current_defaults_to_current(self):
        gitflow = GitFlow(self.repo).init()
        active_branch = self.repo.active_branch
        gitflow.checkout('feature', 'even')
        self.assertNotEqual(gitflow.repo.active_branch.name, active_branch)
        self.assertEqual(gitflow.repo.active_branch.name, 'feat/even')


class TestGitFlowBranches(TestCase):

    #-- nameprefix_or_current

    @copy_from_fixture('sample_repo')
    def test_gitflow_nameprefix_or_current_defaults_to_current(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertEqual(gitflow.nameprefix_or_current('feature', ''),
                         'recursion')

    @copy_from_fixture('sample_repo')
    def test_gitflow_nameprefix_or_current_on_other_branch_type_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertRaises(NoSuchBranchError,
                          gitflow.nameprefix_or_current, 'release', '')

    @copy_from_fixture('sample_repo')
    def test_gitflow_nameprefix_or_current_expands_prefix(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertEqual(gitflow.nameprefix_or_current('feature', 'e'), 'even')

    @copy_from_fixture('sample_repo')
    def test_gitflow_nameprefix_or_current_returns_name(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertEqual('even',
            gitflow.name_or_current('feature', 'even'))

    #-- name_or_current

    @copy_from_fixture('sample_repo')
    def test_gitflow_name_or_current_defaults_to_current(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertEqual(gitflow.name_or_current('feature', ''),
                         'recursion')

    @copy_from_fixture('sample_repo')
    def test_gitflow_name_or_current_on_other_branch_type_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertRaises(NoSuchBranchError,
                          gitflow.name_or_current, 'release', '')

    @copy_from_fixture('sample_repo')
    def test_gitflow_name_or_current_for_nonexisting_name_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertRaises(NoSuchBranchError,
                          gitflow.name_or_current, 'feature', 'xxxx')

    @copy_from_fixture('sample_repo')
    def test_gitflow_name_or_current_returns_name(self):
        gitflow = GitFlow(self.repo).init()
        # gitflow.init checks out `devel` branch :-(
        self.repo.branches['feat/recursion'].checkout()
        self.assertEqual('even',
            gitflow.name_or_current('feature', 'even'))
        self.assertEqual('xxxx',
            gitflow.name_or_current('feature', 'xxxx', must_exist=False))


class TestGitFlowCommandCreate(TestCase):

    # more tests are done in test_branches.py

    def test_create_in_new_sandbox(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox).init()
        gitflow.create('feature', 'wow-feature', base=None, fetch=False)
        self.assertIn('feature/wow-feature', gitflow.repo.branches)

class TestGitFlowCommandFinish(TestCase):

    # more tests are done in test_branches.py

    def test_finish_in_new_sandbox(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox).init()
        gitflow.create('feature', 'wow-feature', base=None, fetch=False)
        self.assertEqual(gitflow.repo.active_branch.name, 'feature/wow-feature')
        fake_commit(gitflow.repo, 'Yet another commit')
        gitflow.finish('feature', 'wow-feature', False, False, False, False, None)
        self.assertNotIn('feature/wow-feature', gitflow.repo.branches)

    def test_finish_in_new_sandbox_without_commit(self):
        sandbox = create_sandbox(self)
        gitflow = GitFlow(sandbox).init()
        gitflow.create('feature', 'wow-feature', base=None, fetch=False)
        gitflow.finish('feature', 'wow-feature', False, False, False, False, None)
        self.assertNotIn('feature/wow-feature', gitflow.repo.branches)

    @copy_from_fixture('sample_repo')
    def test_finish_unresolved_merge_conflict(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.finish('feature', 'recursion', False, False, False, False, None)
        self.assertRaises(MergeError,
                          gitflow.finish, 'feature', 'even', False, False, False, False, None)
        # do not resolve, but finish again
        self.assertRaises(Usage,
                          gitflow.finish, 'feature', 'even', False, False, False, False, None)


class TestGitFlowCommandTrack(TestCase):

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_track_creates_tracking_branch(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.track('feature', 'even')
        self.assertEqual(self.repo.active_branch.name, 'feat/even')
        self.assertTrue(self.repo.active_branch.tracking_branch())
        self.assertEqual(self.repo.refs['feat/even'].tracking_branch(),
                         gitflow.origin().refs['feat/even'])

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_track_existing_branch_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        self.assertRaises(BranchExistsError,
                          gitflow.track, 'feature', 'recursion')

    # :todo: more test-cases for GitFlow.track()


class TestGitFlowCommandRebase(TestCase):

    @copy_from_fixture('sample_repo')
    def test_gitflow_rebase(self):
        gitflow = GitFlow(self.repo).init()
        dc0 = self.repo.refs['devel'].commit
        fc0 = self.repo.refs['feat/even'].commit
        b0 = self.repo.git.merge_base(dc0, fc0)

        gitflow.develop().checkout()
        fake_commit(self.repo, 'A commit on devel')
        dc1 = self.repo.refs['devel'].commit
        b1 = self.repo.git.merge_base(dc1, fc0)
        # commit advances `devel`
        self.assertNotEqual(dc0, dc1)
        # merge base is still the same
        self.assertEqual(b0, b1)

        gitflow.rebase('feature', 'even', interactive=False)

        fc1 = self.repo.refs['feat/even'].commit
        b2 = self.repo.git.merge_base('devel', 'feat/even')
        # rebase advances `feat/even`
        self.assertNotEqual(fc0, fc1)
        # merge base is now new `devel` head
        self.assertEqual(b2, dc1.hexsha)


class TestGitFlowCommandPull(TestCase):

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_creates_non_tracking_branch(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.pull('feature', 'my-remote', 'even')
        self.assertEqual(self.repo.active_branch.name, 'feat/even')
        self.assertEqual(self.repo.active_branch.tracking_branch(), None)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_while_on_other_branchtype_is_allowed(self):
        gitflow = GitFlow(self.repo).init()
        # activate some hotfix branch
        new_branch = self.repo.create_head('hf-dummy', 'stable')
        new_branch.checkout()
        gitflow.pull('feature', 'my-remote', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_while_on_same_branchtype_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        # activate some feature branch
        new_branch = self.repo.create_head('feat/something', 'stable')
        new_branch.checkout()
        # try to pull another feature branch
        self.assertRaisesRegexp(
            SystemExit, "To avoid unintended merges, git-flow aborted.",
            gitflow.pull, 'feature', 'my-remote', 'my-name-is-irrelevant')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_existing_branch_creates_non_tracking_branch(self):
        gitflow = GitFlow(self.repo).init()
        # create local branch based on first commit
        new_branch = self.repo.create_head('feat/even', 'stable')
        new_branch.checkout()
        gitflow.pull('feature', 'my-remote', 'even')
        self.assertEqual(self.repo.active_branch.name, 'feat/even')
        self.assertEqual(self.repo.active_branch.tracking_branch(), None)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_existing_branch_while_on_other_branchtype_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        # create local branch based on first commit
        new_branch = self.repo.create_head('feat/even', 'stable')
        self.assertRaisesRegexp(
            SystemExit, "To avoid unintended merges, git-flow aborted.",
            gitflow.pull, 'feature', 'my-remote', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_existing_branch_while_on_same_branchtype_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        # create local branch based on first commit
        new_branch = self.repo.create_head('feat/even', 'stable')
        # activate some feature branch
        new_branch = self.repo.create_head('feat/something', 'stable')
        new_branch.checkout()
        self.assertRaisesRegexp(
            SystemExit, "To avoid unintended merges, git-flow aborted.",
            gitflow.pull, 'feature', 'my-remote', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_non_existing_feature_raises_error(self):
        all_remote_commits_before_change = all_commits(self.remote)
        gitflow = GitFlow(self.repo).init()
        self.assertRaisesRegexp(
            GitCommandError, "Couldn't find remote ref ",
            gitflow.pull, 'feature', 'my-remote', 'i-am-not-here')

    # :todo: pull changed and test if new commit is really pulled
    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_pull_really_pulls(self):
        gitflow = GitFlow(self.repo).init()
        self.remote.heads['feat/even'].checkout()
        change = fake_commit(self.remote, "Another commit")
        self.assertNotIn(change, all_commits(self.repo))
        gitflow.pull('feature', 'my-remote', 'even')
        self.assertIn(change, all_commits(self.repo))

    # :todo: pull_requires_clean_working_tree


class TestGitFlowCommandPublish(TestCase):

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_creates_remote_branch(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.create('feature', 'circular', 'devel', fetch=False)
        gitflow.publish('feature', 'circular')
        self.assertIn('feat/circular', self.remote.branches)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_creates_sets_tracking_branch(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.create('feature', 'circular', 'devel', fetch=False)
        gitflow.publish('feature', 'circular')

        self.assertTrue(self.repo.branches['feat/circular'].tracking_branch())
        self.assertTrue(self.repo.branches['feat/circular'].tracking_branch().name,
                        'my-remote/feat/circular')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_really_pushes(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.create('feature', 'circular', 'devel', fetch=False)
        change = fake_commit(self.repo, "Another commit")
        all_local_commits = all_commits(self.repo)
        self.assertIn(change, all_local_commits)
        gitflow.publish('feature', 'circular')

        all_remote_commits = all_commits(self.remote)
        self.assertEqual(all_remote_commits, all_remote_commits)
        self.assertIn(change, all_remote_commits)

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_non_existing_branch_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        self.assertRaises(NoSuchBranchError,
                          gitflow.publish, 'feature', 'new-feature')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_branch_existing_remote_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        self.assertRaises(NoSuchBranchError,
                          gitflow.publish, 'feature', 'even')

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_existing_remote_branch_raises_error(self):
        gitflow = GitFlow(self.repo).init()
        gitflow.create('feature', 'even', 'devel', fetch=False)
        self.assertRaises(BranchExistsError,
                          gitflow.publish, 'feature', 'even')

    # :todo: publish_requires_clean_working_tree


class TestGitFlowCommandDiff(TestCase):

    @remote_clone_from_fixture('sample_repo')
    def test_gitflow_publish_creates_remote_branch(self):
        gitflow = GitFlow(self.repo).init()
        orig_stdout = sys.stdout
        sys.stdout = StringIO.StringIO()
        gitflow.diff('feature', 'recursion')
        diff = sys.stdout.getvalue()
        sys.stdout = orig_stdout
        difflines = diff.splitlines()
        matchlines = [
            'diff --git a/odd.py b/odd.py',
            'index 607a269..8a0c7ff 100644',
            '--- a/odd.py',
            '+++ b/odd.py',
            ]
        self.assertEqual(difflines[:len(matchlines)], matchlines)


class TestGitFlowBranchManagement(TestCase):

    # Branch type detection
    def test_detect_branch_types(self):
        create_git_repo(self)
        gitflow = GitFlow()

        # The types that "ship" with git-flow
        self.assertIn('feature', gitflow.managers)
        self.assertIn('release', gitflow.managers)
        self.assertIn('hotfix', gitflow.managers)
        self.assertIn('support', gitflow.managers)

    def test_detect_custom_branch_types(self):
        create_git_repo(self)
        # Declare a custom branch type inline
        class FooBarManager(BranchManager):
            identifier = 'foobar'
            DEFAULT_PREFIX = 'xyz/'

        gitflow = GitFlow()
        self.assertIn('foobar', gitflow.managers)

    def test_empty_branch_type_list_raises_usage(self):
        create_git_repo(self)
        gitflow = GitFlow().init()
        self.assertRaises(Usage, gitflow.list, 'feature', 'name', False, False)
        self.assertRaises(Usage, gitflow.list, 'release', 'version', False, False)
        self.assertRaises(Usage, gitflow.list, 'hotfix', 'version', False, False)
        self.assertRaises(Usage, gitflow.list, 'support', 'version', False, False)

    # Branch creation
    def test_create_branches(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        gitflow.create('feature', 'foo', None, fetch=False)
        self.assertIn('feature/foo',
                [h.name for h in gitflow.repo.branches])
        gitflow.create('release', '1.0', None, fetch=False)
        self.assertIn('release/1.0',
                [h.name for h in gitflow.repo.branches])

    def _test_create_branches_from_alt_base(self):
        repo = create_git_repo(self)
        gitflow = GitFlow(repo).init()
        gitflow.create('feature', 'foo', 'master', fetch=False)
        self.assertIn('feature/foo',
                [h.name for h in gitflow.repo.branches])
        gitflow.repo.index.commit('Foo')
        gitflow.create('release', '1.0', 'feature/foo', fetch=False)
        self.assertIn('release/1.0',
                [h.name for h in gitflow.repo.branches])


    """
    Use case:

    $ git flow init
    $ git flow feature start foo
    $ git flow status
      develop: 4826bdf
      master: 4826bdf
    * feature/foo: 7c8928a
    $ git commit -m 'foo'
    $ git commit -m 'bar'
    $ git flow feature start bar
    $ git commit -m 'qux'
    $ git flow feature finish foo
    $ git flow undo        # new!
    $ git flow feature finish bar
    There were merge conflicts, resolve them now!
    $ git flow abort       # new!

    """

